import os
import shutil

import pytest
from pkgcore.fs import contents, fs, livefs, ops
from snakeoil.data_source import local_source
from snakeoil.osutils import pjoin
from snakeoil.test import TestCase
from snakeoil.test.mixins import TempDirMixin


class VerifyMixin:

    def verify(self, obj, kwds, stat):
        for attr, keyword in (("st_mtime", "mtime"),
                              ("st_gid", "gid"),
                              ("st_uid", "uid")):
            if keyword in kwds:
                self.assertEqual(getattr(stat, attr), kwds[keyword],
                                 f"testing {keyword}")
        if "mode" in kwds:
            self.assertEqual((stat.st_mode & 0o4777), kwds["mode"])


class TestDefaultEnsurePerms(VerifyMixin, TempDirMixin, TestCase):

    def common_bits(self, creator_func, kls):
        kwds = {"mtime":0o1234, "uid":os.getuid(), "gid":os.getgid(),
                "mode":0o775, "dev":None, "inode":None}
        o = kls(pjoin(self.dir, "blah"), **kwds)
        creator_func(o.location)
        self.assertTrue(ops.ensure_perms(o))
        self.verify(o, kwds, os.stat(o.location))
        kwds["mode"] = 0o770
        o2 = kls(pjoin(self.dir, "blah"), **kwds)
        self.assertTrue(ops.ensure_perms(o2))
        self.verify(o2, kwds, os.stat(o.location))
        self.assertRaises(
            OSError,
            ops.ensure_perms, kls(pjoin(self.dir, "asdf"), **kwds))

    def test_dir(self):
        self.common_bits(os.mkdir, fs.fsDir)

    def test_file(self):
        self.common_bits(lambda s:open(s, "w").close(), fs.fsFile)


class TestDefaultMkdir(TempDirMixin, TestCase):

    def test_it(self):
        o = fs.fsDir(pjoin(self.dir, "mkdir_test"), strict=False)
        self.assertTrue(ops.mkdir(o))
        old_umask = os.umask(0)
        try:
            self.assertEqual((os.stat(o.location).st_mode & 0o4777), 0o777 & ~old_umask)
        finally:
            os.umask(old_umask)
        os.rmdir(o.location)
        o = fs.fsDir(pjoin(self.dir, "mkdir_test2"), strict=False, mode=0o750)
        self.assertTrue(ops.mkdir(o))
        self.assertEqual(os.stat(o.location).st_mode & 0o4777, 0o750)


class TestCopyFile(VerifyMixin, TempDirMixin, TestCase):

    def test_it(self):
        src = pjoin(self.dir, "copy_test_src")
        dest = pjoin(self.dir, "copy_test_dest")
        with open(src, "w") as f:
            f.writelines("asdf\n" for i in range(10))
        kwds = {"mtime":10321, "uid":os.getuid(), "gid":os.getgid(),
                "mode":0o664, "data":local_source(src), "dev":None,
                "inode":None}
        o = fs.fsFile(dest, **kwds)
        self.assertTrue(ops.copyfile(o))
        with open(dest, "r") as f:
            self.assertEqual("asdf\n" * 10, f.read())
        self.verify(o, kwds, os.stat(o.location))

    def test_sym_perms(self):
        curgid = os.getgid()
        group = [x for x in os.getgroups() if x != curgid]
        if not group or os.getuid() != 0:
            pytest.skip(
                "requires root privs for this test, or for this user to"
                "belong to more then one group"
            )
        group = group[0]
        fp = pjoin(self.dir, "sym")
        o = fs.fsSymlink(fp, mtime=10321, uid=os.getuid(), gid=group,
            mode=0o664, target='target')
        self.assertTrue(ops.copyfile(o))
        self.assertEqual(os.lstat(fp).st_gid, group)
        self.assertEqual(os.lstat(fp).st_uid, os.getuid())

    def test_puke_on_dirs(self):
        path = pjoin(self.dir, "puke_dir")
        self.assertRaises(TypeError,
            ops.copyfile,
            fs.fsDir(path, strict=False))
        os.mkdir(path)
        fp = pjoin(self.dir, "foon")
        open(fp, "w").close()
        f = livefs.gen_obj(fp)
        self.assertRaises(TypeError,
            livefs.gen_obj(fp).change_attributes(location=path))

        # test sym over a directory.
        f = fs.fsSymlink(path, fp, mode=0o644, mtime=0, uid=os.getuid(),
            gid=os.getgid())
        self.assertRaises(TypeError, ops.copyfile, f)
        os.unlink(fp)
        os.mkdir(fp)
        self.assertRaises(ops.CannotOverwrite, ops.copyfile, f)


class ContentsMixin(VerifyMixin, TempDirMixin, TestCase):

    entries_norm1 = {
        "file1":["reg"],
        "dir":["dir"],
        "dir/subdir":["dir"],
        "dir/file2":["reg"],
        "ldir":["sym", "dir/subdir"],
        "dir/lfile":["sym", "../file1"]
        }

    entries_rec1 = {
        "dir":["dir"],
        "dir/link":["sym", "../dir"]
        }

    def generate_tree(self, base, entries):
        s_ents = [(pjoin(base, k), entries[k]) for k in sorted(entries)]
        for k, v in s_ents:
            if v[0] == "dir":
                os.mkdir(k)
        for k, v in s_ents:
            if v[0] == "dir":
                pass
            elif v[0] == "reg":
                open(k, "w").close()
            elif v[0] == "sym":
                os.symlink(v[1], k)
            else:
                raise Exception(
                    f"generate_tree doesn't support type {v!r} yet: k {k!r}")

    def gen_dir(self, name):
        d = os.path.join(self.dir, name)
        if os.path.exists(d):
            shutil.rmtree(d)
        os.mkdir(d)
        return d


class Test_merge_contents(ContentsMixin):

    def generic_merge_bits(self, entries):
        src = self.gen_dir("src")
        self.generate_tree(src, entries)
        cset = livefs.scan(src, offset=src)
        dest = self.gen_dir("dest")
        self.assertTrue(ops.merge_contents(cset, offset=dest))
        self.assertEqual(livefs.scan(src, offset=src),
            livefs.scan(dest, offset=dest))
        return src, dest, cset

    def test_callback(self):
        for attr in dir(self):
            if not attr.startswith('entries') or 'fail' in attr:
                continue
            e = getattr(self, attr)
            if not isinstance(e, dict):
                continue
            src, dest, cset = self.generic_merge_bits(e)
            new_cset = contents.contentsSet(contents.offset_rewriter(dest, cset))
            s = set(new_cset)
            ops.merge_contents(cset, offset=dest, callback=s.remove)
            self.assertFalse(s)

    def test_dangling_symlink(self):
        src = self.gen_dir("src")
        self.generate_tree(src, {"dir":["dir"]})
        cset = livefs.scan(src, offset=src)
        dest = self.gen_dir("dest")
        os.symlink(pjoin(dest, "dest"), pjoin(dest, "dir"))
        self.assertTrue(ops.merge_contents(cset, offset=dest))
        self.assertEqual(cset, livefs.scan(src, offset=dest))

    def test_empty_overwrite(self):
        self.generic_merge_bits(self.entries_norm1)

    def test_recursive_links(self):
        self.generic_merge_bits(self.entries_rec1)

    def test_exact_overwrite(self):
        src, dest, cset = self.generic_merge_bits(self.entries_norm1)
        self.assertTrue(ops.merge_contents(cset, offset=dest))

    def test_sym_over_dir(self):
        path = pjoin(self.dir, "sym")
        fp = pjoin(self.dir, "trg")
        os.mkdir(path)
        # test sym over a directory.
        f = fs.fsSymlink(path, fp, mode=0o644, mtime=0, uid=os.getuid(),
            gid=os.getgid())
        cset = contents.contentsSet([f])
        self.assertRaises(ops.FailedCopy, ops.merge_contents, cset)
        self.assertTrue(fs.isdir(livefs.gen_obj(path)))
        os.mkdir(fp)
        ops.merge_contents(cset)

    def test_dir_over_file(self):
        # according to the spec, dirs can't be merged over files that
        # aren't dirs or symlinks to dirs
        path = pjoin(self.dir, "file2dir")
        open(path, 'w').close()
        d = fs.fsDir(path, mode=0o755, mtime=0, uid=os.getuid(), gid=os.getgid())
        cset = contents.contentsSet([d])
        self.assertRaises(ops.CannotOverwrite, ops.merge_contents, cset)


class Test_unmerge_contents(ContentsMixin):

    def generic_unmerge_bits(self, entries, img="img"):
        img = self.gen_dir(img)
        self.generate_tree(img, entries)
        cset = livefs.scan(img, offset=img)
        return img, cset

    def test_callback(self):
        for attr in dir(self):
            if not attr.startswith('entries') or 'fail' in attr:
                continue
            e = getattr(self, attr)
            if not isinstance(e, dict):
                continue
            img, cset = self.generic_unmerge_bits(e)
            s = set(contents.offset_rewriter(img, cset))
            ops.unmerge_contents(cset, offset=img, callback=s.remove)
            self.assertFalse(s, s)

    def test_empty_removal(self):
        img, cset = self.generic_unmerge_bits(self.entries_norm1)
        self.assertTrue(
            ops.unmerge_contents(cset, offset=os.path.join(self.dir, "dest")))

    def test_exact_removal(self):
        img, cset = self.generic_unmerge_bits(self.entries_norm1)
        self.assertTrue(ops.unmerge_contents(cset, offset=img))
        self.assertFalse(livefs.scan(img, offset=img))

    def test_lingering_file(self):
        img, cset = self.generic_unmerge_bits(self.entries_norm1)
        dirs = [k for k, v in self.entries_norm1.items() if v[0] == "dir"]
        fp = os.path.join(img, dirs[0], "linger")
        open(fp, "w").close()
        self.assertTrue(ops.unmerge_contents(cset, offset=img))
        self.assertTrue(os.path.exists(fp))



__all__ = ("UnknownData", "FixLibtoolArchivesTrigger")

import re
from functools import partial
from os.path import basename, dirname

from snakeoil.currying import post_curry
from snakeoil.sequences import stable_unique

from ..exceptions import PkgcoreException
from ..merge import triggers

x11_sub = post_curry(partial(
    re.compile(r"X11R6/+lib").sub, "lib"), 1)
local_sub = post_curry(partial(
    re.compile(r"local/+lib").sub, "lib"), 1)
pkgconfig1_sub = post_curry(partial(
    re.compile(r"usr/+lib[^/]*/+pkgconfig/+\.\./\.\.").sub,
    "usr"), 1)
pkgconfig2_sub = post_curry(partial(
    re.compile(r"usr/+lib[^/]*/+pkgconfig/+\.\.").sub,
    "usr"), 1)
flags_match = re.compile(
    r"-(?:mt|mthreads|kthread|Kthread|pthread"
    r"|pthreads|-thread-safe|threads)").match

template = """# %(file)s - a libtool library file
# Generated by ltmain.sh - GNU libtool 1.5.10 (1.1220.2.130 2004/09/19 12:13:49)
%(content)s
"""


class UnknownData(PkgcoreException):

    def __init__(self, line, token=None):
        self.token, self.line = token, line

    def __str__(self):
        s = f"we don't know how to parse line {self.line!r}"
        if self.token:
            s += f"specifically token {self.token!r}"
        return s

# libtiff.la - a libtool library file
# Generated by ltmain.sh - GNU libtool 1.5.10 (1.1220.2.130 2004/09/19 12:13:49)
def parse_lafile(handle):
    d = {}
    for line in handle:
        line = line.rstrip()
        if not line or line.startswith("#"):
            continue
        try:
            key, val = line.split("=", 1)
        except ValueError:
            raise UnknownData(line)
        # ' is specifically forced by libtdl implementation
        if len(val) >= 2 and val[0] == val[-1] and val[0] == "'":
            val = val[1:-1]
        d[key] = val
    return d


def rewrite_lafile(handle, filename):
    data = parse_lafile(handle)
    raw_dep_libs = data.get("dependency_libs", False)
    if not raw_dep_libs:
        return False, None

    original_libs = raw_dep_libs.split()
    rpaths, libs, libladirs, inherited_flags = [], [], [], []
    original_inherited_flags = data.get("inherited_linker_flags", [])

    for item in stable_unique(original_libs):
        if item.startswith("-l"):
            libs.append(item)
        elif item.endswith(".la"):
            base = basename(item)
            if base.startswith("lib"):
                # convert to -l; punt .la, and 'lib' prefix
                libs.append("-l" + base[3:-3])
                libladirs.append("-L" + dirname(item))
            else:
                libs.append(item)
        elif item.startswith("-L"):
                # this is heinous, but is what the script did.
                item = x11_sub(item)
                item = local_sub(item)
                item = pkgconfig1_sub(item)
                item = pkgconfig2_sub(item)
                libladirs.append(item)
        elif item.startswith("-R"):
            rpaths.append(item)
        elif flags_match(item):
            if inherited_flags:
                inherited_flags.append(item)
            else:
                libs.append(item)
        else:
            raise UnknownData(raw_dep_libs, item)
    libs = stable_unique(rpaths + libladirs + libs)
    inherited_flags = stable_unique(inherited_flags)
    if libs == original_libs and inherited_flags == original_inherited_flags:
        return False, None

    # must be prefixed with a space
    data["dependency_libs"] = ' ' + (' '.join(libs))
    if inherited_flags:
        # must be prefixed with a space
        data["inherited_flags"] = ' ' + (' '.join(inherited_flags))
    content = "\n".join(f"{k}='{v}'" for k,v in sorted(data.items()))
    return True, template % {"content":content, "file":filename}

def fix_fsobject(location):
    from ..fs import fs, livefs
    for obj in livefs.iter_scan(location):
        if not fs.isreg(obj) or not obj.basename.endswith(".la"):
            continue

        with open(obj.location, 'r') as f:
            updated, content = rewrite_lafile(f, obj.basename)
        if updated:
            with open(obj.location, 'w') as f:
                f.write(content)


class FixLibtoolArchivesTrigger(triggers.base):

    required_csets = ('install',)
    _engine_types = triggers.INSTALLING_MODES
    _hooks = ('pre_merge',)

    def trigger(self, engine, cset):
        updates = []
        for obj in cset.iterfiles():
            if not obj.basename.endswith(".la"):
                continue
            handle = obj.data.text_fileobj()
            updated, content = rewrite_lafile(handle, obj.basename)
            if not updated:
                continue

            engine.observer.info(f"rewriting libtool archive {obj.location}")
            source = engine.get_writable_fsobj(obj, empty=True)
            source.text_fileobj(True).write(content)
            # force chksums to be regenerated
            updates.append(obj.change_attributes(data=source,
                chksums=None))
        cset.update(updates)



from pkgcore import log
from pkgcore.restrictions import packages, values
from pkgcore.test import TestCase, TestRestriction, callback_logger, malleable_obj, silence_logging
from snakeoil.mappings import AttrAccessible


class AlwaysSelfIntersect(values.base):
    def intersect(self, other):
        return self

    __hash__ = object.__hash__


class TestPackageRestriction(TestRestriction):

    if packages.PackageRestriction is packages.PackageRestriction:
        kls = packages.PackageRestriction
    else:
        class kls(packages.PackageRestriction,
            packages.PackageRestriction_mixin):
            __slots__ = ()
            __inst_caching__ = packages.PackageRestriction.__inst_caching__

    kls = staticmethod(kls)

    @silence_logging(log.logging.root)
    def test_matching(self):
        strexact = values.StrExactMatch

        args = malleable_obj(category="foon", package="dar")
        self.assertMatches(self.kls("category", strexact("foon")), args)
        self.assertMatches(self.kls("package", strexact("dar")), args)
        self.assertNotMatches(self.kls("package", strexact("dar"), negate=True),
            args)
        self.assertNotMatches(self.kls("package", strexact("foon")), args)

        self.assertMatches(self.kls("package", strexact("foon"), negate=True),
            args)
        excepts = []
        # no msg should be thrown, it wasn't an unexpected exception

        log.logging.root.addHandler(callback_logger(excepts.append))
        self.assertNotMatches(self.kls("foon", AlwaysSelfIntersect), args)
        self.assertFalse(excepts)

        self.assertMatches(self.kls("foon", AlwaysSelfIntersect, negate=True),
            args)
        self.assertFalse(excepts)

        class foo:
            def __getattr__(self, attr):
                if attr.startswith("exc"):
                    raise exceptions_d.get(attr[4:], None)()
                raise AttributeError("monkey lover")

        exceptions_d = {"KeyboardInterrupt":KeyboardInterrupt,
            "RuntimeError":RuntimeError, "SystemExit":SystemExit}

        for mode in ("match", "force_True", "force_False"):
            excepts[:] = []
            self.assertRaises(AttributeError,
                getattr(self.kls("foon", AlwaysSelfIntersect), mode),
                foo())
            self.assertEqual(
                len(excepts), 1,
                msg=f"expected one exception, got {excepts!r}")

            # ensure various exceptions are passed through
            for k in (KeyboardInterrupt, RuntimeError, SystemExit):
                self.assertRaises(
                    k,
                    getattr(self.kls(f"exc_{k.__name__}", AlwaysSelfIntersect), mode),
                    foo())

        # check that it only does string comparison in exception catching.
        class foo:
            def __cmp__(self, other):
                raise TypeError

            def __getattr__(self, attr):
                raise AttributeError(self, attr)

        self.assertFalse(self.kls("foon", AlwaysSelfIntersect).match(foo()))

    def test_attr(self):
        self.assertEqual(self.kls('val', values.AlwaysTrue).attr,
            'val')
        self.assertEqual(self.kls('val.dar', values.AlwaysTrue).attr,
            'val.dar')
        self.assertEqual(self.kls('val', values.AlwaysTrue).attrs,
            ('val',))
        self.assertEqual(self.kls('val.dar', values.AlwaysTrue).attrs,
            ('val.dar',))

    def test_eq(self):
        self.assertEqual(
            self.kls('one', values.AlwaysTrue),
            self.kls('one', values.AlwaysTrue))
        self.assertNotEqual(
            self.kls('one', values.AlwaysTrue),
            self.kls('one', values.AlwaysTrue, negate=True))
        self.assertNotEqual(
            self.kls('one', values.AlwaysTrue),
            self.kls('two', values.AlwaysTrue))
        self.assertNotEqual(
            self.kls('one', values.AlwaysTrue, negate=True),
            self.kls('one', values.AlwaysFalse, negate=True))

    def test_hash(self):
        inst = self.kls('one.dar', AlwaysSelfIntersect())
        hash(inst)


class values_callback(values.base):

    __slots__ = ("callback",)

    def __init__(self, callback):
        object.__setattr__(self, 'callback', callback)

    def match(self, val):
        return self.callback((None, val))

    def force_True(self, pkg, attr, val):
        return self.callback((True, pkg, attr, val))

    def force_False(self, pkg, attr, val):
        return self.callback((False, pkg, attr, val))


class TestPackageRestrictionMulti(TestCase):

    if packages.PackageRestriction is packages.PackageRestriction:
        kls = packages.PackageRestrictionMulti
    else:
        class kls(packages.PackageRestrictionMulti,
            packages.PackageRestrictionMulti_mixin):
            __slots__ = ()
            __inst_caching__ = packages.PackageRestrictionMulti.__inst_caching__

    kls = staticmethod(kls)

    def test_attr(self):
        o = self.kls(("asdf.far", "repo"), values.AlwaysTrue)
        self.assertEqual(o.attrs, ("asdf.far", "repo"))
        self.assertEqual(o.attr, None)

    def test_values(self):
        l = []
        def f(*args):
            self.assertLen(args, 1)
            l.append(args[0])
            return True

        o = self.kls(("asdf.far", "repo"), values_callback(f))

        pkg = AttrAccessible()
        o.match(pkg)
        self.assertFalse(l)

        pkg['repo'] = 1
        o.match(pkg)
        self.assertFalse(l)

        pkg['asdf'] = AttrAccessible(far=2)
        o.match(pkg)
        self.assertEqual(l, [(None, [2,1],)])

        l[:] = []
        o.force_True(pkg)
        self.assertEqual(l, [(True, pkg, ('asdf.far', 'repo'), [2,1],)])

        l[:] = []
        o.force_False(pkg)
        self.assertEqual(l, [(False, pkg, ('asdf.far', 'repo'), [2,1],)])


class ConditionalTest(TestCase):

    def test_eq(self):
        p = (packages.PackageRestriction('one', values.AlwaysTrue),)
        p2 = (packages.PackageRestriction('one', values.AlwaysFalse),)
        v = values.AlwaysTrue
        v2 = values.AlwaysFalse
        self.assertEqual(
            packages.Conditional('use', v, p),
            packages.Conditional('use', v, p))
        self.assertNotEqual(
            packages.Conditional('use', v2, p),
            packages.Conditional('use', v, p))
        self.assertNotEqual(
            packages.Conditional('use', v, p),
            packages.Conditional('use', v, p2))
        self.assertNotEqual(
            packages.Conditional('use1', v, p),
            packages.Conditional('use', v, p))



"""
binpkg ebuild repository
"""

__all__ = ("tree", "ConfiguredTree", "force_unpacking")

import errno
import os

from snakeoil import chksum, compression
from snakeoil.data_source import data_source, local_source
from snakeoil.klass import alias_attr, jit_attr, jit_attr_named
from snakeoil.mappings import DictMixin, StackedDict
from snakeoil.osutils import access, listdir_dirs, listdir_files, pjoin

from ..config.hint import ConfigHint
from ..ebuild import ebd, ebuild_built
from ..ebuild.cpv import VersionedCPV
from ..fs.contents import contentsSet, offset_rewriter
from ..fs.livefs import scan
from ..fs.ops import merge_contents
from ..fs.tar import generate_contents
from ..merge import engine, triggers
from ..package import base as pkg_base
from ..repository import errors, prototype, wrapper
from . import remote, repo_ops
from .xpak import Xpak


class force_unpacking(triggers.base):

    required_csets = ('new_cset',)
    priority = 5
    _hooks = ('sanity_check',)
    _label = 'forced decompression'
    _engine_type = triggers.INSTALLING_MODES

    def __init__(self, format_op):
        self.format_op = format_op

    def trigger(self, engine, cset):
        op = self.format_op
        op = getattr(op, 'install_op', op)
        op.setup_workdir()
        merge_cset = cset
        if engine.offset != '/':
            merge_cset = cset.change_offset(engine.offset, '/')
        merge_contents(merge_cset, offset=op.env["D"])

        # ok.  they're on disk.
        # now to avoid going back to the binpkg, we rewrite
        # the data_source for files to the on disk location.
        # we can update in place also, since we're not changing the mapping.

        # this rewrites the data_source to the ${D} loc.
        d = op.env["D"]
        fi = (x.change_attributes(data=local_source(
            pjoin(d, x.location.lstrip('/'))))
            for x in merge_cset.iterfiles())

        if engine.offset:
            # we're using merge_cset above, which has the final offset loc
            # pruned; this is required for the merge, however, we're updating
            # the cset so we have to insert the final offset back in.
            # wrap the iter, iow.
            fi = offset_rewriter(engine.offset, fi)

        cset.update(contentsSet(fi))

        # we *probably* should change the csets class at some point
        # since it no longer needs to be tar, but that's for another day.
        engine.replace_cset('new_cset', cset)


class BinPkg(ebuild_built.generate_new_factory):

    def _add_format_triggers(self, pkg, op_inst, format_op_inst,
                                engine_inst):
        if (engine.UNINSTALL_MODE != engine_inst.mode and
                pkg == engine_inst.new and pkg.repo is engine_inst.new.repo):
            t = force_unpacking(op_inst.format_op)
            t.register(engine_inst)

        klass._add_format_triggers(
            self, pkg, op_inst, format_op_inst, engine_inst)

    def scan_contents(self, location):
        return scan(location, offset=location)


class StackedXpakDict(DictMixin):
    __slots__ = ("_xpak", "_parent", "_pkg", "contents", "_wipes", "_chf_obj")

    _metadata_rewrites = {
        "bdepend": "BDEPEND",
        "depend": "DEPEND",
        "rdepend": "RDEPEND",
        "pdepend": "PDEPEND",
        "idepend": "IDEPEND",
        "use": "USE",
        "eapi": "EAPI",
        "CONTENTS": "contents",
        "fullslot": "SLOT",
    }

    def __init__(self, parent, pkg):
        self._pkg = pkg
        self._parent = parent
        self._wipes = set()

    @jit_attr
    def xpak(self):
        return Xpak(self._parent._get_path(self._pkg))

    mtime = alias_attr('_chf_.mtime')

    @jit_attr_named('_chf_obj')
    def _chf_(self):
        return chksum.LazilyHashedPath(self._parent._get_path(self._pkg))

    def __getitem__(self, key):
        key = self._metadata_rewrites.get(key, key)
        if key in self._wipes:
            raise KeyError(self, key)
        if key == "contents":
            data = generate_contents(self._parent._get_path(self._pkg))
            object.__setattr__(self, "contents", data)
        elif key == "environment":
            data = self.xpak.get("environment.bz2")
            if data is None:
                data = data_source(self.xpak.get("environment"), mutable=True)
                if data is None:
                    raise KeyError(
                        "environment.bz2 not found in xpak segment, "
                        "malformed binpkg?")
            else:
                data = data_source(
                    compression.decompress_data('bzip2', data), mutable=True)
        elif key == "ebuild":
            data = self.xpak.get(f"{self._pkg.package}-{self._pkg.fullver}.ebuild", "")
            data = data_source(data)
        else:
            try:
                data = self.xpak[key]
            except KeyError:
                if key == '_eclasses_':
                    # hack...
                    data = {}
                else:
                    data = ''
        return data

    def __delitem__(self, key):
        if key in ("contents", "environment"):
            if key in self._wipes:
                raise KeyError(self, key)
            self._wipes.add(key)
        else:
            del self.xpak[key]

    def __setitem__(self, key, val):
        if key in ("contents", "environment"):
            setattr(self, key, val)
            self._wipes.discard(key)
        else:
            self.xpak[key] = val
        return val

    def keys(self):
        for k in self.xpak:
            yield k
        for k in ("environment", "contents"):
            if self.get(k) is not None:
                yield k

    def __contains__(self, key):
        translated_key = self._metadata_rewrites.get(key, key)
        if translated_key in self._wipes:
            return False
        elif key in ('ebuild', 'environment', 'contents'):
            return True
        return translated_key in self.xpak


class StackedCache(StackedDict):

    __externally_mutable__ = True

    def __delitem__(self, key):
        self._dicts[0].pop(key)


class tree(prototype.tree):

    # yes, the period is required. no, do not try and remove it
    # (harring says it stays)
    extension = ".tbz2"

    configured = False
    configurables = ("settings",)
    operations_kls = repo_ops.operations
    cache_name = "Packages"

    pkgcore_config_type = ConfigHint({
        'location': 'str',
        'repo_id': 'str'},
        typename='repo')

    def __init__(self, location, repo_id=None, cache_version='0'):
        """
        :param location: root of the tbz2 repository
        :keyword repo_id: unique repository id to use; else defaults to
            the location
        """
        super().__init__()
        self.base = self.location = location
        if repo_id is None:
            repo_id = location
        self.repo_id = repo_id
        self._versions_tmp_cache = {}

        # XXX rewrite this when snakeoil.osutils grows an access equivalent.
        if not access(self.base, os.X_OK | os.R_OK):
            # either it doesn't exist, or we don't have perms.
            if not os.path.exists(self.base):
                raise errors.InitializationError(f"base {self.base!r} doesn't exist")
            raise errors.InitializationError(
                "base directory %r with mode 0%03o isn't readable/executable"
                " by this user" %
                (self.base, os.stat(self.base).st_mode & 0o4777))

        self.cache = remote.get_cache_kls(cache_version)(pjoin(self.base, self.cache_name))
        self.package_class = BinPkg(self)

    def configure(self, *args):
        return(ConfiguredTree(self, *args))

    def __str__(self):
        return self.repo_id

    def _get_categories(self, *optional_category):
        # return if optional_category is passed... cause it's not yet supported
        if optional_category:
            return {}
        try:
            return tuple(
                x for x in listdir_dirs(self.base)
                if x.lower() != "all")
        except EnvironmentError as e:
            raise KeyError(f"failed fetching categories: {e}") from e

    def _get_packages(self, category):
        cpath = pjoin(self.base, category.lstrip(os.path.sep))
        l = set()
        d = {}
        lext = len(self.extension)
        bad = False
        try:
            for x in listdir_files(cpath):
                # don't use lstat; symlinks may exist
                if (x.endswith(".lockfile") or
                        not x[-lext:].lower() == self.extension or
                        x.startswith(".tmp.")):
                    continue
                pv = x[:-lext]
                pkg = VersionedCPV(f'{category}/{pv}')
                l.add(pkg.package)
                d.setdefault((category, pkg.package), []).append(pkg.fullver)
        except EnvironmentError as e:
            raise KeyError(
                "failed fetching packages for category %s: %s" %
                (pjoin(self.base, category.lstrip(os.path.sep)), str(e))) from e

        self._versions_tmp_cache.update(d)
        return tuple(l)

    def _get_versions(self, catpkg):
        return tuple(self._versions_tmp_cache.pop(catpkg))

    def _get_path(self, pkg):
        return pjoin(self.base, pkg.category, f"{pkg.package}-{pkg.fullver}.tbz2")

    _get_ebuild_path = _get_path

    def _get_metadata(self, pkg, force=False):
        xpak = StackedXpakDict(self, pkg)
        try:
            if force:
                raise KeyError
            cache_data = self.cache[pkg.cpvstr]
            if int(cache_data['mtime']) != int(xpak.mtime):
                raise KeyError
        except KeyError:
            cache_data = self.cache.update_from_xpak(pkg, xpak)
        obj = StackedCache(cache_data, xpak)
        return obj

    def notify_add_package(self, pkg):
        prototype.tree.notify_add_package(self, pkg)
        # XXX horrible hack.
        self._get_metadata(self.match(pkg.versioned_atom)[0], force=True)
        self.cache.commit()

    def notify_remove_package(self, pkg):
        prototype.tree.notify_remove_package(self, pkg)
        try:
            os.rmdir(pjoin(self.base, pkg.category))
        except OSError as oe:
            # POSIX specifies either ENOTEMPTY or EEXIST for non-empty dir
            # in particular, Solaris uses EEXIST in that case.
            # https://github.com/pkgcore/pkgcore/pull/181
            if oe.errno not in (errno.ENOTEMPTY, errno.EEXIST):
                raise
            del oe

    @property
    def _repo_ops(self):
        return repo_ops


class _WrappedBinpkg(pkg_base.wrapper):
    """Binary package with configuration data bound to it."""

    built = True
    __slots__ = ()

    def __str__(self):
        return (
            f'ebuild binary pkg: {self.cpvstr}::{self.repo.repo_id}, '
            f'source repo {self.source_repository!r}'
        )


class ConfiguredTree(wrapper.tree):
    """Configured repository for portage-compatible binary packages."""

    configured = True

    def __init__(self, repo, domain_settings):
        _WrappedBinpkg._operations = self._generate_operations
        _WrappedBinpkg.repo = self
        super().__init__(repo, package_class=_WrappedBinpkg)
        self.domain_settings = domain_settings

    def _generate_operations(self, domain, pkg, **kwargs):
        return ebd.built_operations(
            domain, pkg._raw_pkg, initial_env=self.domain_settings, **kwargs)



import sys
from io import StringIO

from pkgcore.config import central, cparser, errors
from snakeoil.test import TestCase


class CaseSensitiveConfigParserTest(TestCase):

    def test_case_sensitivity(self):
        cp = cparser.CaseSensitiveConfigParser()
        config = StringIO('\n'.join((
            '[header]',
            'foo=bar',
            'FOO=BAR',
            '[HEADER]',
            'foo=notbar',
            )))
        if sys.hexversion < 0x03020000:
            cp.readfp(config)
        else:
            cp.read_file(config)
        self.assertEqual(cp.get('header', 'foo'), 'bar')
        self.assertEqual(cp.get('header', 'FOO'), 'BAR')
        self.assertEqual(cp.get('HEADER', 'foo'), 'notbar')


class ConfigFromIniTest(TestCase):

    def test_config_from_ini(self):
        config = cparser.config_from_file(StringIO('''
[test]
string = 'hi I am a string'
list = foo bar baz
list.prepend = pre bits
list.append = post bits
true = yes
false = no
'''))
        self.assertEqual(list(config.keys()), ['test'])
        section = config['test']
        for key, arg_type, value in [
            ('string', 'str', [None, 'hi I am a string', None]),
            ('list', 'list', [
                    ['pre', 'bits'], ['foo', 'bar', 'baz'], ['post', 'bits']]),
            ('true', 'bool', True),
            ('false', 'bool', False),
            ]:
            self.assertEqual(section.render_value(None, key, arg_type), value)

    def test_missing_section_ref(self):
        config = cparser.config_from_file(StringIO('''
[test]
ref = 'missing'
'''))
        section = config['test']
        self.assertRaises(
            errors.ConfigurationError,
            section.render_value(
                central.ConfigManager([]), 'ref', 'ref:drawer').collapse)



"""
package class for buildable ebuilds
"""

__all__ = (
    "Maintainer", "MetadataXml", "LocalMetadataXml",
    "SharedPkgData", "Licenses", "OverlayedProfiles",
    "Project", "ProjectMember", "Subproject", "ProjectsXml", "LocalProjectsXml"
)

import contextlib
import errno
import os
import platform
import subprocess
from collections import namedtuple
from itertools import chain
from sys import intern

from lxml import etree
from snakeoil import klass, mappings
from snakeoil.bash import BashParseError, iter_read_bash, read_dict
from snakeoil.caching import WeakInstMeta
from snakeoil.currying import post_curry
from snakeoil.fileutils import readfile, readlines
from snakeoil.osutils import listdir, listdir_files, pjoin
from snakeoil.osutils.mount import umount
from snakeoil.process.namespaces import simple_unshare
from snakeoil.sequences import iter_stable_unique
from snakeoil.strings import pluralism

from ..config.hint import ConfigHint
from ..exceptions import PermissionDenied
from ..log import logger
from ..repository import errors as repo_errors
from ..repository import syncable
from ..restrictions import packages
from . import atom, pkg_updates, profiles
from .eapi import get_eapi


class Maintainer:
    """Data on a single maintainer.

    At least one of email and name is not C{None}.

    :type email: C{unicode} object or C{None}
    :ivar email: email address.
    :type name: C{unicode} object or C{None}
    :ivar name: full name
    :type description: C{unicode} object or C{None}
    :ivar description: description of maintainership.
    :type maint_type: C{unicode} object or C{None}
    :ivar maint_type: maintainer type (person or project).
    :type proxied: C{unicode} object or C{None}
    :ivar proxied: proxied maintainer status (yes, no, proxy)
    """

    __slots__ = ('email', 'description', 'name', 'maint_type', 'proxied')

    def __init__(self, email=None, name=None, description=None, maint_type=None, proxied=None):
        if email is None and name is None:
            raise ValueError('need at least one of name and email')
        self.email = email
        self.name = name
        self.description = description
        self.maint_type = maint_type
        self.proxied = proxied

    def __str__(self):
        if self.name is not None:
            if self.email is not None:
                res = f'{self.name} <{self.email}>'
            else:
                res = self.name
        else:
            res = self.email
        if self.description is not None:
            return f'{res} ({self.description})'
        return res

    def __eq__(self, other):
        try:
            return self.email == other.email and self.name == other.name
        except AttributeError:
            if isinstance(other, str):
                return other == self.email or other == self.name
        return False

    def __hash__(self):
        return hash((self.email, self.name))


class Upstream:
    """Data on a single upstream."""

    __slots__ = ('type', 'name')

    def __init__(self, type, name):
        self.type = type
        self.name = name

    def __eq__(self, other):
        try:
            return self.type == other.type and self.name == other.name
        except AttributeError:
            pass
        return False

    def __hash__(self):
        return hash((self.type, self.name))


class MetadataXml:
    """metadata.xml parsed results

    Attributes are set to -1 if unloaded, None if no entry, or the value
    if loaded.
    """

    __slots__ = (
        "__weakref__", "_maintainers", "_upstreams", "_local_use",
        "_longdescription", "_source", "_stabilize_allarches",
    )

    def __init__(self, source):
        self._source = source

    def _generic_attr(self, attr):
        if self._source is not None:
            self._parse_xml()
        return getattr(self, attr)

    for attr in ("maintainers", "upstreams", "local_use",
                 "longdescription", "stabilize_allarches"):
        locals()[attr] = property(post_curry(_generic_attr, "_" + attr))
    del attr

    def _parse_xml(self, source=None):
        if source is None:
            source = self._source.bytes_fileobj()
        try:
            tree = etree.parse(source)
        except etree.XMLSyntaxError as e:
            self._maintainers = ()
            self._upstreams = ()
            self._local_use = mappings.ImmutableDict()
            self._longdescription = None
            self._source = None
            self._stabilize_allarches = False
            logger.error(e)
            return

        # TODO: handle i18n properly
        maintainers = []
        for x in tree.findall("maintainer"):
            name = email = description = None
            for e in x:
                if e.tag == "name":
                    name = e.text
                elif e.tag == "email":
                    email = e.text
                elif e.tag == 'description' and e.get('lang', 'en') == 'en':
                    description = e.text
            try:
                maintainers.append(Maintainer(
                    name=name, email=email, description=description,
                    maint_type=x.get('type'), proxied=x.get('proxied')))
            except ValueError:
                # ignore invalid maintainers that should be caught by pkgcheck
                pass

        self._maintainers = tuple(maintainers)
        self._upstreams = tuple(
            Upstream(e.get('type'), e.text) for e in chain.from_iterable(tree.findall("upstream"))
            if e.tag == 'remote-id'
        )

        # Could be unicode!
        self._longdescription = None
        for x in tree.findall("longdescription"):
            if x.get('lang', 'en') != 'en':
                continue
            longdesc = ''.join(x.itertext())
            if longdesc:
                self._longdescription = ' '.join(longdesc.split())
            break

        self._source = None

        # lang="" is property of <use/>
        self._local_use = mappings.ImmutableDict()
        for x in tree.findall("use"):
            if x.get('lang', 'en') != 'en':
                continue
            self._local_use = mappings.ImmutableDict(
                (e.attrib['name'], ' '.join(''.join(e.itertext()).split()))
                for e in x.findall('flag')
                if 'name' in e.attrib
            )
            break

        self._stabilize_allarches = tree.find("stabilize-allarches") is not None


class LocalMetadataXml(MetadataXml):

    __slots__ = ()

    def _parse_xml(self):
        try:
            with open(self._source, "rb", 32768) as src:
                MetadataXml._parse_xml(self, src)
        except FileNotFoundError:
            self._maintainers = ()
            self._upstreams = ()
            self._local_use = mappings.ImmutableDict()
            self._longdescription = None
            self._source = None
            self._stabilize_allarches = False


class SharedPkgData:

    __slots__ = ("__weakref__", "metadata_xml", "manifest")

    def __init__(self, metadata_xml, manifest):
        self.metadata_xml = metadata_xml
        self.manifest = manifest


class ProjectMember(metaclass=klass.generic_equality):
    """Data on a single project member.

    :type email: C{unicode} object
    :ivar email: email address.
    :type name: C{unicode} object or C{None}
    :ivar name: full name
    :type role: C{unicode} object or C{None}
    :ivar role: role within the project.
    :type is_lead: C{bool}
    :ivar is_lead: whether the member is a project lead.
    """

    __slots__ = ('email', 'name', 'role', 'is_lead')
    __attr_comparison__ = ('email', 'name', 'role', 'is_lead')

    def __init__(self, email, name=None, role=None, is_lead=None):
        if email is None:
            raise ValueError('email for project member must not be null')
        self.email = email
        self.name = name
        self.role = role
        self.is_lead = is_lead

    def __str__(self):
        if self.name is not None:
            res = f'{self.name} <{self.email}>'
        else:
            res = self.email
        if self.role is not None:
            return f'{res} ({self.role})'
        return res


class Subproject:
    """Data on a subproject.

    :type inherit_members: C{bool}
    :ivar inherit_members: whether the parent project inherits members from this subproject
    """

    __slots__ = ('_ref', 'inherit_members', '_projects_xml', '_project')

    def __init__(self, ref, projects_xml, inherit_members=None):
        if ref is None:
            raise ValueError('ref for subproject must not be null')
        self._ref = ref
        self.inherit_members = inherit_members
        self._projects_xml = projects_xml

    @klass.jit_attr
    def project(self):
        try:
            return self._projects_xml.projects[self._ref]
        except KeyError:
            logger.error(f'projects.xml: subproject {self._ref!r} does not exist')
            return None

    __getattr__ = klass.GetAttrProxy('project')
    __dir__ = klass.DirProxy('project')


class Project:
    """Data on a single project.

    :type email: C{unicode} object
    :ivar email: email address.
    :type name: C{unicode} object or C{None}
    :ivar name: full name
    :type url: C{unicode} object or C{None}
    :ivar url: project website URI
    :type description: C{unicode} object or C{None}
    :ivar description: full project description.
    :type members: C{tuple} of C{ProjectMember}
    :ivar members: project members
    :type subprojects: C{tuple} of C{Subprojects}
    :ivar subprojects: subprojects
    """

    __slots__ = ('email', 'name', 'url', 'description', 'members', 'subprojects')

    def __init__(self, email, name=None, url=None, description=None,
                 members=(), subprojects=()):
        if email is None:
            raise ValueError('email for project must not be null')
        self.email = email
        self.name = name
        self.url = url
        self.description = description
        self.members = tuple(members)
        self.subprojects = tuple(subprojects)

    def __str__(self):
        if self.name is not None:
            res = f'{self.name} <{self.email}>'
        else:
            res = self.email
        if self.url is not None:
            return f'{res} ({self.url})'
        return res

    @property
    def leads(self):
        """Project lead(s), if any."""
        return tuple(m for m in self.members if m.is_lead)

    @property
    def recursive_members(self):
        """All project members, including members inherited from subprojects."""
        subprojects = list(
            sp for sp in self.subprojects
            if sp.inherit_members and sp.project is not None)
        subproject_emails = set(sp.email for sp in subprojects)

        # recursively collect all subprojects from which to inherit
        i = 0
        while i < len(subprojects):
            for sp in subprojects[i].subprojects:
                if sp.project is None:
                    continue
                if sp.inherit_members and sp.email not in subproject_emails:
                    subprojects.append(sp)
                    subproject_emails.add(sp.email)
            i += 1

        members = {m.email: m for m in self.members}
        for sp in (x for x in subprojects if x.inherit_members):
            for m in sp.members:
                if m.email not in members:
                    # drop lead bit
                    m = ProjectMember(
                        email=m.email, name=m.name, role=m.role, is_lead=False)
                    members[m.email] = m
        return tuple(members.values())


class ProjectsXml:
    """projects.xml parsed results

    Attributes are set to -1 if unloaded, None if no entry, or the value
    if loaded.
    """

    __slots__ = ('__weakref__', '_projects', '_source')

    def __init__(self, source):
        self._source = source

    @klass.jit_attr
    def projects(self):
        if self._source is not None:
            return self._parse_xml()
        return mappings.ImmutableDict()

    def _parse_xml(self, source=None):
        if source is None:
            source = self._source.bytes_fileobj()
        try:
            tree = etree.parse(source)
        except etree.XMLSyntaxError as e:
            logger.error(f'failed parsing projects.xml: {e}')
            return mappings.ImmutableDict()

        projects = {}
        for p in tree.findall('project'):
            kwargs = {}
            for k in ('email', 'name', 'url', 'description'):
                kwargs[k] = p.findtext(k)

            members = []
            for m in p.findall('member'):
                m_kwargs = {}
                for k in ('email', 'name', 'role'):
                    m_kwargs[k] = m.findtext(k)
                m_kwargs['is_lead'] = m.get('is-lead', '') == '1'
                try:
                    members.append(ProjectMember(**m_kwargs))
                except ValueError:
                    logger.error(f"project {kwargs['email']} has <member/> with no email")
            kwargs['members'] = members

            subprojects = []
            for sp in p.findall('subproject'):
                try:
                    subprojects.append(Subproject(
                        ref=sp.get('ref'),
                        inherit_members=sp.get('inherit-members', '') == '1',
                        projects_xml=self))
                except ValueError:
                    logger.error(f"project {kwargs['email']} has <subproject/> with no ref")
            kwargs['subprojects'] = subprojects

            projects[kwargs['email']] = Project(**kwargs)

        return mappings.ImmutableDict(projects)


class LocalProjectsXml(ProjectsXml):

    __slots__ = ()

    def _parse_xml(self):
        try:
            with open(self._source, "rb", 32768) as f:
                return super()._parse_xml(f)
        except FileNotFoundError:
            return mappings.ImmutableDict()


class Licenses(metaclass=WeakInstMeta):

    __inst_caching__ = True
    __slots__ = ('_base', '_licenses', '_groups', 'license_groups_path', 'licenses_dir', '_repo_masters', '_license_instances')

    def __init__(self, repo, *repo_masters,
                 licenses_dir='licenses', license_groups='profiles/license_groups'):
        repo_base = repo.location
        object.__setattr__(self, '_base', repo_base)
        object.__setattr__(self, 'license_groups_path', pjoin(repo_base, license_groups))
        object.__setattr__(self, 'licenses_dir', pjoin(repo_base, licenses_dir))
        object.__setattr__(self, '_repo_masters', repo_masters)
        self._load_license_instances()

    def _load_license_instances(self):
        l = []
        for x in self._repo_masters:
            if isinstance(x, Licenses):
                l.append(x)
            elif hasattr(x, 'licenses'):
                l.append(x.licenses)
        object.__setattr__(self, '_license_instances', tuple(l))

    @klass.jit_attr_none
    def licenses(self):
        """Return the set of all defined licenses in a repo."""
        try:
            content = listdir_files(self.licenses_dir)
        except EnvironmentError:
            content = ()
        return frozenset(chain(content, *self._license_instances))

    @klass.jit_attr_none
    def groups(self):
        """Return the mapping of defined license groups to licenses for a repo."""
        try:
            d = read_dict(self.license_groups_path, splitter=' ')
            for k, v in d.items():
                d[k] = set(v.split())
        except EnvironmentError:
            return mappings.ImmutableDict()
        except BashParseError as pe:
            logger.error(f"failed parsing license_groups: {pe}")
            return mappings.ImmutableDict()
        for li in self._license_instances:
            for k, v in li.groups.items():
                if k in d:
                    d[k] |= v
                else:
                    d[k] = v
        self._expand_groups(d)
        return mappings.ImmutableDict((k, frozenset(v)) for (k, v) in d.items())

    def _expand_groups(self, groups):
        keep_going = True
        while keep_going:
            keep_going = False
            for k, v in groups.items():
                if not any(x[0] == '@' for x in v):
                    continue
                keep_going = True
                l = []
                for v2 in v:
                    if v2[0] == '@':
                        v2 = v2[1:]
                        if not v2 or v2 not in groups:
                            logger.error(
                                f"invalid license group reference: {v2!r} in {self}")
                            continue
                        elif v2 == k:
                            logger.error(
                                f"cyclic license group references for {v2!r} in {self}")
                            continue
                        l.extend(groups[v2])
                    else:
                        l.append(v2)
                groups[k] = l

    def refresh(self):
        self._load_license_instances()
        for li in self._license_instances:
            li.refresh()
        self._licenses = None
        self._groups = None

    def __getitem__(self, license):
        if license not in self:
            for li in self._license_instances:
                with contextlib.suppress(KeyError):
                    return li[license]
            raise KeyError(license)
        try:
            return open(pjoin(self.licenses_dir, license)).read()
        except FileNotFoundError:
            raise KeyError(license)

    def __len__(self):
        return len(self.licenses)

    def __iter__(self):
        return iter(self.licenses)

    def __contains__(self, license):
        return license in self.licenses


class _immutable_attr_dict(mappings.ImmutableDict):

    __slots__ = ()

    mappings.inject_getitem_as_getattr(locals())


_KnownProfile = namedtuple('_KnownProfile', ['base', 'arch', 'path', 'status', 'deprecated'])


class Profiles(klass.ImmutableInstance):

    __slots__ = ('config', 'profiles_base', '_profiles')
    __inst_caching__ = True

    def __init__(self, repo_config, profiles_base=None):
        object.__setattr__(self, 'config', repo_config)
        profiles_base = profiles_base if profiles_base is not None else repo_config.profiles_base
        object.__setattr__(self, 'profiles_base', profiles_base)

    @klass.jit_attr_none
    def profiles(self):
        return self.parse(self.profiles_base, self.config.repo_id)

    @staticmethod
    def parse(profiles_base, repo_id, known_status=None, known_arch=None):
        """Return the mapping of arches to profiles for a repo."""
        l = []
        fp = pjoin(profiles_base, 'profiles.desc')
        try:
            for lineno, line in iter_read_bash(fp, enum_line=True):
                try:
                    arch, profile, status = line.split()
                except ValueError:
                    logger.error(
                        f"{repo_id}::profiles/profiles.desc, "
                        f"line {lineno}: invalid profile line format: "
                        "should be 'arch profile status'")
                    continue
                if known_status is not None and status not in known_status:
                    logger.warning(
                        f"{repo_id}::profiles/profiles.desc, "
                        f"line {lineno}: unknown profile status: {status!r}")
                if known_arch is not None and arch not in known_arch:
                    logger.warning(
                        f"{repo_id}::profiles/profiles.desc, "
                        f"line {lineno}: unknown arch: {arch!r}")
                # Normalize the profile name on the offchance someone slipped an extra /
                # into it.
                path = '/'.join(filter(None, profile.split('/')))
                deprecated = os.path.exists(
                    os.path.join(profiles_base, path, 'deprecated'))
                l.append(_KnownProfile(profiles_base, arch, path, status, deprecated))
        except FileNotFoundError:
            # no profiles exist
            pass
        return frozenset(l)

    def __len__(self):
        return len(self.profiles)

    def __iter__(self):
        yield from self.profiles

    def __getitem__(self, path):
        if path[0] == '/':
            path = path.lstrip(self.profiles_base).lstrip(os.sep)
        for p in self.profiles:
            if p.path == path:
                return p
        raise KeyError(path)

    def __contains__(self, path):
        if path[0] == '/':
            path = path.lstrip(self.profiles_base).lstrip(os.sep)
        for p in self.profiles:
            if p.path == path:
                return True
        return False

    def refresh(self):
        self._profiles = None

    def arches(self, status=None):
        """All arches with profiles defined in the repo optionally matching a given status."""
        arches = []
        for p in self.profiles:
            if status is None or status == p.status:
                arches.append(p.arch)
        return frozenset(arches)

    def get_profiles(self, status):
        """Yield profiles matching a given status."""
        for p in self.profiles:
            if status == p.status or (status == 'deprecated' and p.deprecated):
                yield p

    def create_profile(self, node, **kwargs):
        """Return profile object for a given, parsed profile entry."""
        return profiles.OnDiskProfile(node.base, node.path, **kwargs)


class OverlayedProfiles(Profiles):

    __inst_caching__ = True
    __slots__ = ('_profiles_instances', '_profiles_sources')

    def __init__(self, *profiles_sources):
        object.__setattr__(self, '_profiles_sources', profiles_sources)
        self._load_profiles_instances()

    @klass.jit_attr_none
    def profiles(self):
        return frozenset(chain.from_iterable(self._profiles_instances))

    def refresh(self):
        self._load_profiles_instances()
        for pi in self._profiles_instances:
            pi.refresh()
        Profiles.refresh(self)

    def _load_profiles_instances(self):
        l = []
        for x in self._profiles_sources:
            if isinstance(x, Profiles):
                l.append(x)
            elif hasattr(x, 'profiles'):
                l.append(x.profiles)
        object.__setattr__(self, '_profiles_instances', tuple(l))


class RepoConfig(syncable.tree, klass.ImmutableInstance, metaclass=WeakInstMeta):
    """Configuration data for an ebuild repository."""

    layout_offset = "metadata/layout.conf"

    default_hashes = ('size', 'blake2b', 'sha512')
    default_required_hashes = ('size', 'blake2b')
    supported_profile_formats = ('pms', 'portage-1', 'portage-2', 'profile-set')
    supported_cache_formats = ('md5-dict', 'pms')

    __inst_caching__ = True

    pkgcore_config_type = ConfigHint(
        typename='repo_config',
        types={
            'config_name': 'str',
            'syncer': 'lazy_ref:syncer',
        })

    def __init__(self, location, config_name=None, syncer=None, profiles_base='profiles'):
        super().__init__(syncer)
        object.__setattr__(self, 'config_name', config_name)
        object.__setattr__(self, 'external', (config_name is None))
        object.__setattr__(self, 'location', location)
        object.__setattr__(self, 'profiles_base', pjoin(self.location, profiles_base))

        try:
            self._parse_config()
        except OSError as e:
            raise repo_errors.InitializationError(str(e))

        if not self.eapi.is_supported:
            raise repo_errors.UnsupportedRepo(self)

    def _parse_config(self):
        """Load data from the repo's metadata/layout.conf file."""
        path = pjoin(self.location, self.layout_offset)
        data = read_dict(
            iter_read_bash(readlines(path, strip_whitespace=True, swallow_missing=True)),
            source_isiter=True, strip=True, filename=path, ignore_errors=True)

        sf = object.__setattr__
        sf(self, 'repo_name', data.get('repo-name', None))

        hashes = data.get('manifest-hashes', '').lower().split()
        if hashes:
            hashes = ['size'] + hashes
            hashes = tuple(iter_stable_unique(hashes))
        else:
            hashes = self.default_hashes

        required_hashes = data.get('manifest-required-hashes', '').lower().split()
        if required_hashes:
            required_hashes = ['size'] + required_hashes
            required_hashes = tuple(iter_stable_unique(required_hashes))
        else:
            required_hashes = self.default_required_hashes

        manifest_policy = data.get('use-manifests', 'strict').lower()
        d = {
            'disabled': (manifest_policy == 'false'),
            'strict': (manifest_policy == 'strict'),
            'thin': (data.get('thin-manifests', '').lower() == 'true'),
            'signed': (data.get('sign-manifests', 'true').lower() == 'true'),
            'hashes': hashes,
            'required_hashes': required_hashes,
        }

        sf(self, 'manifests', _immutable_attr_dict(d))
        masters = data.get('masters')
        _missing_masters = False
        if masters is None:
            if not self.is_empty:
                logger.warning(
                    f"{self.repo_id} repo at {self.location!r}, doesn't "
                    "specify masters in metadata/layout.conf. Please explicitly "
                    "set masters (use \"masters =\" if the repo is standalone).")
            _missing_masters = True
            masters = ()
        else:
            masters = tuple(iter_stable_unique(masters.split()))
        sf(self, '_missing_masters', _missing_masters)
        sf(self, 'masters', masters)
        aliases = data.get('aliases', '').split() + [
            self.config_name, self.repo_name, self.pms_repo_name, self.location]
        sf(self, 'aliases', tuple(filter(None, iter_stable_unique(aliases))))
        sf(self, 'eapis_deprecated', tuple(iter_stable_unique(data.get('eapis-deprecated', '').split())))
        sf(self, 'eapis_banned', tuple(iter_stable_unique(data.get('eapis-banned', '').split())))
        sf(self, 'eapis_testing', tuple(iter_stable_unique(data.get('eapis-testing', '').split())))
        sf(self, 'profile_eapis_deprecated', tuple(iter_stable_unique(data.get('profile-eapis-deprecated', '').split())))
        sf(self, 'profile_eapis_banned', tuple(iter_stable_unique(data.get('profile-eapis-banned', '').split())))
        sf(self, 'properties_allowed', tuple(iter_stable_unique(data.get('properties-allowed', '').split())))
        sf(self, 'restrict_allowed', tuple(iter_stable_unique(data.get('restrict-allowed', '').split())))
        sf(self, 'sign_commits', data.get('sign-commits', 'false').lower() == 'true')

        v = set(data.get('cache-formats', 'md5-dict').lower().split())
        if not v:
            v = [None]
        else:
            # sort into favored order
            v = [f for f in self.supported_cache_formats if f in v]
            if not v:
                logger.warning('unknown cache format: falling back to md5-dict format')
                v = ['md5-dict']
        sf(self, 'cache_format', list(v)[0])

        profile_formats = set(data.get('profile-formats', 'pms').lower().split())
        if not profile_formats:
            logger.info(
                f"{self.repo_id!r} repo at {self.location!r} has explicitly "
                "unset profile-formats, defaulting to pms")
            profile_formats = {'pms'}
        unknown = profile_formats.difference(self.supported_profile_formats)
        if unknown:
            logger.info(
                "%r repo at %r has unsupported profile format%s: %s",
                self.repo_id, self.location, pluralism(unknown),
                ', '.join(sorted(unknown)))
            profile_formats.difference_update(unknown)
            profile_formats.add('pms')
        sf(self, 'profile_formats', profile_formats)

    @klass.jit_attr
    def known_arches(self):
        """All valid KEYWORDS for the repo."""
        try:
            return frozenset(iter_read_bash(
                pjoin(self.profiles_base, 'arch.list')))
        except FileNotFoundError:
            return frozenset()

    @klass.jit_attr
    def arches_desc(self):
        """Arch stability status (GLEP 72).

        See https://www.gentoo.org/glep/glep-0072.html for more details.
        """
        fp = pjoin(self.profiles_base, 'arches.desc')
        d = {'stable': set(), 'transitional': set(), 'testing': set()}
        try:
            for lineno, line in iter_read_bash(fp, enum_line=True):
                try:
                    arch, status = line.split()
                except ValueError:
                    logger.error(
                        f"{self.repo_id}::profiles/arches.desc, "
                        f"line {lineno}: invalid line format: "
                        "should be '<arch> <status>'")
                    continue
                if arch not in self.known_arches:
                    logger.warning(
                        f"{self.repo_id}::profiles/arches.desc, "
                        f"line {lineno}: unknown arch: {arch!r}")
                    continue
                if status not in d:
                    logger.warning(
                        f"{self.repo_id}::profiles/arches.desc, "
                        f"line {lineno}: unknown status: {status!r}")
                    continue
                d[status].add(arch)
        except FileNotFoundError:
            pass
        return mappings.ImmutableDict(d)

    @klass.jit_attr
    def use_desc(self):
        """Global USE flags for the repo."""
        # todo: convert this to using a common exception base, with
        # conversion of ValueErrors...
        def converter(key):
            return (packages.AlwaysTrue, key)
        return tuple(self._split_use_desc_file('use.desc', converter))

    @klass.jit_attr
    def use_local_desc(self):
        """Local USE flags for the repo."""
        def converter(key):
            # todo: convert this to using a common exception base, with
            # conversion of ValueErrors/atom exceptions...
            chunks = key.split(':', 1)
            return (atom.atom(chunks[0]), chunks[1])

        return tuple(self._split_use_desc_file('use.local.desc', converter))

    @klass.jit_attr
    def use_expand_desc(self):
        """USE_EXPAND settings for the repo."""
        base = pjoin(self.profiles_base, 'desc')
        d = {}
        try:
            targets = listdir_files(base)
        except FileNotFoundError:
            targets = []

        for use_group in targets:
            group = use_group.split('.', 1)[0]
            d[group] = tuple(
                self._split_use_desc_file(
                    f'desc/{use_group}', lambda k: f'{group}_{k}', matcher=False))

        return mappings.ImmutableDict(d)

    @klass.jit_attr
    def use_expand_sort(self):
        """Mapping of USE_EXPAND sorting keys for the repo."""
        base = pjoin(self.profiles_base, 'desc')
        d = {}
        try:
            targets = listdir_files(base)
        except FileNotFoundError:
            targets = []

        for use_group in targets:
            group = use_group.split('.', 1)[0]
            use_expand = (x[0] for x in self._split_use_desc_file(
                f'desc/{use_group}', lambda k: k, matcher=False))
            d[group] = {use: i for i, use in enumerate(use_expand)}

        return mappings.ImmutableDict(d)

    def _split_use_desc_file(self, name, converter, matcher=True):
        line = None
        fp = pjoin(self.profiles_base, name)
        try:
            for line in iter_read_bash(fp):
                try:
                    key, val = line.split(None, 1)
                    key = converter(key)
                    if matcher:
                        yield key[0], (key[1], val.split('-', 1)[1].strip())
                    else:
                        yield key, val.split('-', 1)[1].strip()
                except ValueError as e:
                    logger.error(f'failed parsing {fp!r}, line {line!r}: {e}')
        except FileNotFoundError:
            pass
        except ValueError as e:
            logger.error(f'failed parsing {fp!r}: {e}')

    @klass.jit_attr
    def is_empty(self):
        """Return boolean related to if the repo has files in it."""
        result = True
        try:
            # any files existing means it's not empty
            result = not listdir(self.location)
            if result:
                logger.debug(f"repo is empty: {self.location!r}")
        except FileNotFoundError:
            pass

        return result

    @klass.jit_attr
    def pms_repo_name(self):
        """Repository name from profiles/repo_name (as defined by PMS).

        We're more lenient than the spec and don't verify it conforms to the
        specified format.
        """
        name = readfile(pjoin(self.profiles_base, 'repo_name'), none_on_missing=True)
        if name is not None:
            name = name.split('\n', 1)[0].strip()
        return name

    @klass.jit_attr
    def repo_id(self):
        """Main identifier for the repo.

        The precedence order is as follows: repos.conf name, repo-name from
        metadata/layout.conf, profiles/repo_name, and finally a fallback to the
        repo's location for unlabeled repos.
        """
        if self.config_name:
            return self.config_name
        # repo_name might not be parsed yet if failure occurs during init
        if repo_name := getattr(self, 'repo_name', None):
            return repo_name
        if self.pms_repo_name:
            return self.pms_repo_name
        if not self.is_empty:
            logger.warning(f"repo lacks a defined name: {self.location!r}")
        return self.location

    @klass.jit_attr
    def updates(self):
        """Package updates for the repo defined in profiles/updates/*."""
        updates_dir = pjoin(self.profiles_base, 'updates')
        d = pkg_updates.read_updates(updates_dir, eapi=self.eapi)
        return mappings.ImmutableDict(d)

    @klass.jit_attr
    def categories(self):
        categories = readlines(pjoin(self.profiles_base, 'categories'), True, True, True)
        if categories is not None:
            return tuple(map(intern, categories))
        return ()

    @klass.jit_attr
    def profiles(self):
        return Profiles(self)

    @klass.jit_attr
    def base_profile(self):
        pms_strict = 'pms' in self.profile_formats
        return profiles.EmptyRootNode(self.profiles_base, pms_strict=pms_strict)

    @klass.jit_attr
    def eapi(self):
        try:
            return self.base_profile.eapi
        except profiles.NonexistentProfile:
            return get_eapi('0')

    @klass.jit_attr
    def pkg_masks(self):
        """Package masks from profiles/package.mask."""
        return frozenset(self.base_profile.masks[1])

    @klass.jit_attr
    def pkg_deprecated(self):
        """Deprecated packages from profiles/package.deprecated."""
        return frozenset(self.base_profile.pkg_deprecated[1])


class SquashfsRepoConfig(RepoConfig):
    """Configuration data for an ebuild repository in a squashfs archive.

    Linux only support for transparently supporting ebuild repos compressed
    into squashfs archives.
    """

    def __init__(self, sqfs_file, location, *args, **kwargs):
        sqfs_path = pjoin(location, sqfs_file)
        object.__setattr__(self, '_sqfs', sqfs_path)
        object.__setattr__(self, 'location', location)
        # if squashfs archive exists in the repo, try to mount it over itself
        if os.path.exists(self._sqfs):
            try:
                self._mount_archive()
            except PermissionError:
                if platform.uname().release < '4.18':
                    raise repo_errors.InitializationError(
                        'fuse mounts in user namespaces require linux >= 4.18')
                raise
        super().__init__(location, *args, **kwargs)

    def _pre_sync(self):
        if os.path.ismount(self.location):
            try:
                self._umount_archive()
            except repo_errors.InitializationError:
                pass

    def _post_sync(self):
        if os.path.exists(self._sqfs):
            self._mount_archive()

    def _failed_cmd(self, process, action):
        if process.returncode:
            stderr = process.stderr.decode().strip().lower()
            msg = f'failed {action} squashfs archive: {stderr}'
            if process.returncode == 1:
                raise PermissionDenied(self._sqfs, msg)
            else:
                raise repo_errors.InitializationError(msg)

    def _mount_archive(self):
        """Mount the squashfs archive onto the repo in a mount namespace."""
        # enable a user namespace if not running as root
        unshare_kwds = {'mount': True, 'user': not os.getuid() == 0}
        try:
            simple_unshare(**unshare_kwds)
        except OSError as e:
            raise repo_errors.InitializationError(
                f'namespace support unavailable: {e.strerror}')

        # First try using mount binary to automatically handle setting up loop
        # device -- this only works with real root perms since loopback device
        # mounting (losetup) doesn't work in user namespaces.
        p = subprocess.run(['mount', self._sqfs, self.location], capture_output=True)

        if p.returncode == 0:
            return
        elif p.returncode not in (1, 32):
            # fail out if not a permissions issue (regular or loopback failure inside userns)
            self._failed_cmd(p, 'mounting')

        # fallback to using squashfuse
        try:
            p = subprocess.run(
                ['squashfuse', '-o', 'nonempty', self._sqfs, self.location],
                capture_output=True)
        except FileNotFoundError as e:
            raise repo_errors.InitializationError(
                f'failed mounting squashfs archive: {e.filename} required')

        if p.returncode:
            self._failed_cmd(p, 'mounting')

    def _umount_archive(self):
        """Unmount the squashfs archive."""
        try:
            umount(self.location)
            return
        except FileNotFoundError as e:
            raise repo_errors.InitializationError(
                f'failed unmounting squashfs archive: {e.filename} required')
        except OSError as e:
            # fail out if not a permissions issue (regular or loopback failure inside userns)
            if e.errno not in (errno.EPERM, errno.EPIPE):
                raise repo_errors.InitializationError(
                    f'failed unmounting squashfs archive: {e.strerror}')

        # fallback to using fusermount
        try:
            p = subprocess.run(['fusermount', '-u', self.location], capture_output=True)
        except FileNotFoundError as e:
            raise repo_errors.InitializationError(
                f'failed unmounting squashfs archive: {e.filename} required')

        if p.returncode:
            self._failed_cmd(p, 'unmounting')



__all__ = ("file_identifier",)

from snakeoil.compatibility import IGNORED_EXCEPTIONS
from snakeoil.klass import jit_attr
from snakeoil.process.spawn import spawn_get_output


class file_identifier:

    def __init__(self, force_binary=False):
        if force_binary:
            self.func = self._fallback_file

    def __call__(self, obj):
        if not isinstance(obj, str):
            obj = obj.path
        return self.func(obj)

    @jit_attr
    def func(self):
        try:
            import magic
        except ImportError:
            return self._fallback_file
        if hasattr(magic, 'MAGIC_NONE'):
            # <5.05 of file
            magic_const = magic.MAGIC_NONE
        else:
            magic_const = magic.NONE
        try:
            obj = magic.open(magic_const)
            ret = obj.load()
            if ret == 0:
                return obj.file
        except IGNORED_EXCEPTIONS:
            raise
        except Exception:
            pass # POS of library.
        return self._fallback_file

    @staticmethod
    def _fallback_file(path):
        ret, out = spawn_get_output(["file", path])
        if ret != 0:
            raise ValueError(f"file output was non zero- ret:{ret!r} out:{out!r}")
        out = ''.join(out)
        if out.startswith(path):
            out = out[len(path):]
            if out.startswith(":"):
                out = out[1:]
        return out



import os
from unittest import mock

import pytest
from pkgcore.sync import base, hg
from snakeoil.process import CommandNotFound


class TestHgSyncer:

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path):
        self.repo_path = tmp_path / 'repo'

    def test_uri_parse(self):
        assert hg.hg_syncer.parse_uri("hg+http://dar") == "http://dar"
        assert hg.hg_syncer.parse_uri("mercurial+http://dar") == "http://dar"

        with pytest.raises(base.UriError):
            hg.hg_syncer.parse_uri("hg://dar")

        # external binary doesn't exist
        with mock.patch('snakeoil.process.find_binary') as find_binary:
            find_binary.side_effect = CommandNotFound('svn')
            with pytest.raises(base.SyncError):
                hg.hg_syncer(str(self.repo_path), "hg+http://foon.com/dar")

        # fake that the external binary exists
        with mock.patch('snakeoil.process.find_binary') as find_binary:
            find_binary.return_value = 'hg'
            o = hg.hg_syncer(str(self.repo_path), "hg+http://dar")
            assert o.uri == "http://dar"

    @mock.patch('snakeoil.process.spawn.spawn')
    def test_sync(self, spawn):
        uri = 'https://foo/bar'
        with mock.patch('snakeoil.process.find_binary', return_value='hg'):
            syncer = hg.hg_syncer(str(self.repo_path), f'hg+{uri}')

        # initial sync
        syncer.sync()
        assert spawn.call_args[0] == (
            ['hg', 'clone', uri, str(self.repo_path) + os.path.sep],)
        assert spawn.call_args[1]['cwd'] is None
        # repo update
        self.repo_path.mkdir()
        syncer.sync()
        assert spawn.call_args[0] == (['hg', 'pull', '-u', uri],)
        assert spawn.call_args[1]['cwd'] == syncer.basedir



import os

from pkgcore.fs import livefs
from pkgcore.fs.contents import contentsSet
from pkgcore.merge import engine
from snakeoil.osutils import pjoin
from snakeoil.test import TestCase
from snakeoil.test.mixins import tempdir_decorator

from ..fs.fs_util import fsDir, fsFile, fsSymlink
from .util import fake_engine


class fake_pkg:

    def __init__(self, contents, label=None):
        self.label = label
        self.contents = contents

    def __str__(self):
        return f"fake_pkg: {self.label}"


class Test_MergeEngineCsets(TestCase):

    simple_cset = list(fsFile(x) for x in ("/foon", "/usr/dar", "/blah"))
    simple_cset.extend(fsDir(x) for x in ("/usr", "/usr/lib"))
    simple_cset.append(fsSymlink("/usr/lib/blah", "../../blah"))
    simple_cset.append(fsSymlink("/broken-symlink", "dar"))
    simple_cset = contentsSet(simple_cset, mutable=False)

    kls = engine.MergeEngine

    def assertCsetEqual(self, cset1, cset2):
        if not isinstance(cset1, contentsSet):
            cset1 = contentsSet(cset1)
        if not isinstance(cset2, contentsSet):
            cset2 = contentsSet(cset2)
        self.assertEqual(cset1, cset2, reflective=False)

    def assertCsetNotEqual(self, cset1, cset2):
        if not isinstance(cset1, contentsSet):
            cset1 = contentsSet(cset1)
        if not isinstance(cset2, contentsSet):
            cset2 = contentsSet(cset2)
        self.assertNotEqual(cset1, cset2, reflective=False)

    def run_cset(self, target, engine, *args):
        return getattr(self.kls, target)(engine, engine.csets, *args)

    def test_generate_offset_cset(self):
        engine = fake_engine(csets={"new_cset":self.simple_cset},
            offset='/')
        def run(engine, cset):
            return self.run_cset('generate_offset_cset', engine,
                lambda e, c:c[cset])

        self.assertCsetEqual(self.simple_cset, run(engine, 'new_cset'))
        engine.offset = '/foon/'
        run(engine, 'new_cset')
        self.assertCsetEqual(self.simple_cset.insert_offset(engine.offset),
            run(engine, 'new_cset'))

    def test_get_pkg_contents(self):
        new_cset = self.kls.get_pkg_contents(None, None, fake_pkg(self.simple_cset))
        self.assertCsetEqual(self.simple_cset, new_cset)
        # must differ; shouldn't be modifying the original cset
        self.assertNotIdentical(self.simple_cset, new_cset)

    def test_get_remove_cset(self):
        files = contentsSet(self.simple_cset.iterfiles(invert=True))
        engine = fake_engine(csets={'install':files,
            'old_cset':self.simple_cset})
        self.assertCsetEqual(self.simple_cset.iterfiles(),
            self.run_cset('get_remove_cset', engine))

    def test_get_replace_cset(self):
        files = contentsSet(self.simple_cset.iterfiles(invert=True))
        engine = fake_engine(csets={'install':files,
            'old_cset':self.simple_cset})
        self.assertCsetEqual(files,
            self.run_cset('get_replace_cset', engine))

    @tempdir_decorator
    def test_rewrite_awareness(self):
        src = contentsSet(self.simple_cset)
        src.add(fsFile("/usr/lib/donkey"))
        trg = src.difference(["/usr/lib/donkey"])
        trg.add(fsFile("/usr/lib64/donkey"))
        trg = trg.insert_offset(self.dir)
        os.mkdir(pjoin(self.dir, 'usr'))
        os.mkdir(pjoin(self.dir, 'usr', 'lib64'))
        os.symlink('lib64', pjoin(self.dir, 'usr', 'lib'))
        pkg = fake_pkg(src)
        engine = self.kls.install(self.dir, pkg, offset=self.dir)
        result = engine.csets['resolved_install']
        self.assertEqual(sorted(result.iterfiles()), sorted(trg.iterfiles()))

    @tempdir_decorator
    def test_symlink_awareness(self):
        src = contentsSet(self.simple_cset)
        src.add(fsFile("/usr/lib/blah/donkey"))
        trg = src.difference(["/usr/lib/blah/donkey"])
        trg.add(fsFile("/blah/donkey"))
        trg = trg.insert_offset(self.dir)
        pkg = fake_pkg(src)
        engine = self.kls.install(self.dir, pkg, offset=self.dir)
        result = engine.csets['new_cset']
        self.assertEqual(sorted(result.iterfiles()), sorted(trg.iterfiles()))
    test_symlink_awareness.skip = "contentset should handle this"

    @tempdir_decorator
    def test__get_livefs_intersect_cset(self):
        old_cset = self.simple_cset.insert_offset(self.dir)
        # have to add it; scan adds the root node
        old_cset.add(fsDir(self.dir))
        os.mkdir(pjoin(self.dir, "usr"))
        open(pjoin(self.dir, "usr", "dar"), 'w').close()
        open(pjoin(self.dir, 'foon'), 'w').close()
        # note that this *is* a sym in the cset; adding this specific
        # check so that if the code differs, the test breaks, and the tests
        # get updated (additionally, folks may not be aware of the potential)
        open(pjoin(self.dir, 'broken-symlink'), 'w').close()
        engine = fake_engine(csets={'test':old_cset})
        existent = livefs.scan(self.dir)
        generated = self.run_cset('_get_livefs_intersect_cset', engine,
            'test')
        self.assertEqual(generated, existent)



"""
core engine for livefs modifications
"""

__all__ = ("alias_cset", "map_new_cset_livefs", "MergeEngine")

# need better documentation...

# pre merge triggers
# post merge triggers
# ordering?

import io
import operator
import tempfile
import traceback
from functools import partial
from itertools import chain
from multiprocessing import cpu_count

from snakeoil import data_source
from snakeoil.compatibility import IGNORED_EXCEPTIONS
from snakeoil.currying import post_curry
from snakeoil.fileutils import touch
from snakeoil.mappings import ImmutableDict, LazyValDict, StackedDict
from snakeoil.osutils import normpath

from ..fs import contents, livefs
from ..operations import observer as observer_mod
from ..plugin import get_plugins
from . import errors
from .const import INSTALL_MODE, REPLACE_MODE, UNINSTALL_MODE


def alias_cset(alias, engine, csets):
    """alias a cset to another"""
    return csets[alias]


def map_new_cset_livefs(engine, csets, cset_name='new_cset'):
    """Find symlinks on disk that redirect new_cset, and return a livefs localized cset."""
    initial = csets[cset_name]
    ondisk = contents.contentsSet(livefs.intersect(initial.iterdirs(), realpath=False))
    livefs.recursively_fill_syms(ondisk)
    ret = initial.map_directory_structure(ondisk, add_conflicting_sym=True)
    return ret


class MergeEngine:

    install_hooks = {x: [] for x in
        ("sanity_check", "pre_merge", "merge", "post_merge", "final")}
    uninstall_hooks = {x: [] for x in
        ("sanity_check", "pre_unmerge", "unmerge", "post_unmerge", "final")}
    replace_hooks = {x: [] for x in
        set(chain(install_hooks.keys(), uninstall_hooks.keys()))}

    install_csets = {
        "install_existing": "get_install_livefs_intersect",
        "resolved_install": map_new_cset_livefs,
        'new_cset': partial(alias_cset, 'raw_new_cset'),
        "install": partial(alias_cset, 'new_cset'),
        "replace": partial(alias_cset, 'new_cset'),
    }
    uninstall_csets = {
        "uninstall_existing": partial(alias_cset, "uninstall"),
        "uninstall": partial(alias_cset, "old_cset"),
        "old_cset": "get_uninstall_livefs_intersect",
    }
    replace_csets = install_csets.copy()
    replace_csets.update(uninstall_csets)
    replace_csets["modifying"] = (
        lambda e, c: c["resolved_install"].intersection(c["uninstall"]))
    replace_csets["uninstall"] = "get_remove_cset"
    replace_csets["replace"] = "get_replace_cset"
    replace_csets["install_existing"] = "get_install_livefs_intersect"

    install_csets_preserve = ["new_cset"]
    uninstall_csets_preserve = ["old_cset"]
    replace_csets_preserve = ["new_cset", "old_cset"]

    allow_reuse = True

    def __init__(self, mode, tempdir, hooks, csets, preserves, observer,
                 offset=None, disable_plugins=False, parallelism=None):
        if observer is None:
            observer = observer_mod.repo_observer(observer_mod.null_output)
        self.observer = observer
        self.mode = mode
        if tempdir is not None:
            tempdir = normpath(tempdir) + '/'
        self.tempdir = tempdir

        self.parallelism = parallelism if parallelism is not None else cpu_count()
        self.hooks = ImmutableDict((x, []) for x in hooks)

        self.preserve_csets = []
        self.cset_sources = {}
        # instantiate these separately so their values are preserved
        self.preserved_csets = LazyValDict(
            self.preserve_csets, self._get_cset_source)
        for k, v in csets.items():
            if isinstance(v, str):
                v = getattr(self, v, v)
            if not callable(v):
                raise TypeError(
                    "cset values must be either the string name of "
                    f"existing methods, or callables (got {v})")

            if k in preserves:
                self.add_preserved_cset(k, v)
            else:
                self.add_cset(k, v)

        if offset is None:
            offset = "/"
        self.offset = offset

        if not disable_plugins:
            # merge in default triggers first.
            for trigger in get_plugins('triggers'):
                t = trigger()
                t.register(self)

        # merge in overrides
        for hook, triggers in hooks.items():
            for trigger in triggers:
                self.add_trigger(hook, trigger)

        self.regenerate_csets()
        for x in hooks:
            setattr(self, x, partial(self.execute_hook, x))

    @classmethod
    def install(cls, tempdir, pkg, offset=None, observer=None,
                disable_plugins=False):
        """Generate a MergeEngine instance configured for installing a pkg.

        :param tempdir: tempspace for the merger to use; this space it must
            control alone, no sharing.
        :param pkg: :obj:`pkgcore.package.metadata.package` instance to install
        :param offset: any livefs offset to force for modifications
        :param disable_plugins: if enabled, run just the triggers passed in
        :return: :obj:`MergeEngine`
        """
        hooks = {k: [y() for y in v] for (k, v) in cls.install_hooks.items()}

        csets = cls.install_csets.copy()
        if "raw_new_cset" not in csets:
            csets["raw_new_cset"] = post_curry(cls.get_pkg_contents, pkg)
        o = cls(INSTALL_MODE, tempdir, hooks, csets, cls.install_csets_preserve,
                observer, offset=offset, disable_plugins=disable_plugins)

        if o.offset != '/':
            # wrap the results of new_cset to pass through an offset generator
            o.cset_sources["raw_new_cset"] = post_curry(
                o.generate_offset_cset, o.cset_sources["raw_new_cset"])

        o.new = pkg
        return o

    @classmethod
    def uninstall(cls, tempdir, pkg, offset=None, observer=None,
                  disable_plugins=False):
        """Generate a MergeEngine instance configured for uninstalling a pkg.

        :param tempdir: tempspace for the merger to use; this space it must
            control alone, no sharing.
        :param pkg: :obj:`pkgcore.package.metadata.package` instance to uninstall,
            must be from a livefs vdb
        :param offset: any livefs offset to force for modifications
        :param disable_plugins: if enabled, run just the triggers passed in
        :return: :obj:`MergeEngine`
        """
        hooks = {k: [y() for y in v] for (k, v) in cls.uninstall_hooks.items()}
        csets = cls.uninstall_csets.copy()

        if "raw_old_cset" not in csets:
            csets["raw_old_cset"] = post_curry(cls.get_pkg_contents, pkg)
        o = cls(UNINSTALL_MODE, tempdir, hooks, csets, cls.uninstall_csets_preserve,
                observer, offset=offset, disable_plugins=disable_plugins)

        if o.offset != '/':
            # wrap the results of new_cset to pass through an offset generator
            o.cset_sources["old_cset"] = post_curry(
                o.generate_offset_cset, o.cset_sources["old_cset"])

        o.old = pkg
        return o

    @classmethod
    def replace(cls, tempdir, old, new, offset=None, observer=None,
                disable_plugins=False):
        """Generate a MergeEngine instance configured for replacing a pkg.

        :param tempdir: tempspace for the merger to use; this space it must
            control alone, no sharing.
        :param old: :obj:`pkgcore.package.metadata.package` instance to replace,
            must be from a livefs vdb
        :param new: :obj:`pkgcore.package.metadata.package` instance
        :param offset: any livefs offset to force for modifications
        :param disable_plugins: if enabled, run just the triggers passed in
        :return: :obj:`MergeEngine`
        """
        hooks = {k: [y() for y in v] for (k, v) in cls.replace_hooks.items()}

        csets = cls.replace_csets.copy()

        csets.setdefault('raw_old_cset', post_curry(cls.get_pkg_contents, old))
        csets.setdefault('raw_new_cset', post_curry(cls.get_pkg_contents, new))

        o = cls(REPLACE_MODE, tempdir, hooks, csets, cls.replace_csets_preserve,
                observer, offset=offset, disable_plugins=disable_plugins)

        if o.offset != '/':
            for k in ("raw_old_cset", "raw_new_cset"):
                # wrap the results of new_cset to pass through an
                # offset generator
                o.cset_sources[k] = post_curry(
                    o.generate_offset_cset, o.cset_sources[k])

        o.old = old
        o.new = new
        return o

    def replace_cset(self, name, new_cset):
        """Replace the cset referenced by this engine.

        Use only if you know what you're doing.

        :param name: name of the cset
        :new_cset: a contentsSet instance to use
        """
        if name in self.preserved_csets:
            # yes this is evil awareness of LazyValDict internals...
            self.preserved_csets._vals[name] = new_cset
        else:
            raise KeyError(f"attempted to replace a non preserved cset: {name}")

    def regenerate_csets(self):
        """Internal function, reset non preserverd csets.

        Used in transitioning between hook points
        """
        self.csets = StackedDict(self.preserved_csets,
            LazyValDict(self.cset_sources, self._get_cset_source))

    def _get_cset_source(self, key):
        return self.cset_sources[key](self, self.csets)

    def add_preserved_cset(self, cset_name, func):
        """Register a cset generator for use.

        The cset will stay in memory until the engine finishes all steps.

        :param cset_name: what to call the generated cset
        :param func: callable to get the cset
        """
        self.add_cset(cset_name, func)
        self.preserve_csets.append(cset_name)

    def add_cset(self, cset_name, func):
        """Regiser a cset generator for use.

        The cset will be released from memory when it's no longer used.

        :param cset_name: what to call the generated cset
        :param func: callable to get the cset
        """
        if not callable(func):
            raise TypeError("func must be a callable")
        if not isinstance(cset_name, str):
            raise TypeError("cset_name must be a string")
        self.cset_sources[cset_name] = func

    def add_trigger(self, hook_name, trigger, required_csets):
        """Register a :obj:`pkgcore.merge.triggers.base` instance to be executed.

        :param hook_name: engine step to hook the trigger into
        :param trigger: :class:`pkgcore.merge.triggers.base` to add
        """
        if hook_name not in self.hooks:
            raise KeyError(
                f"trigger {trigger!r}'s hook {hook_name} isn't a known hook")

        if required_csets is not None:
            for rcs in required_csets:
                if rcs not in self.cset_sources:
                    if isinstance(rcs, str):
                        raise errors.TriggerUnknownCset(trigger, rcs)

        self.hooks[hook_name].append(trigger)

    def execute_hook(self, hook):
        """Execute any triggers bound to a hook point."""
        try:
            self.phase = hook
            self.regenerate_csets()
            for trigger in sorted(self.hooks[hook], key=operator.attrgetter("priority")):
                # error checking needed here.
                self.observer.trigger_start(hook, trigger)
                try:
                    try:
                        trigger(self, self.csets)
                    except IGNORED_EXCEPTIONS:
                        raise
                    except errors.BlockModification as e:
                        self.observer.error(
                            f"modification was blocked by trigger {trigger!r}: {e}")
                        raise
                    except errors.ModificationError as e:
                        self.observer.error(
                            f"modification error occurred during trigger {trigger!r}: {e}")
                        raise
                    except Exception as e:
                        if not trigger.suppress_exceptions:
                            raise

                        handle = io.StringIO()
                        traceback.print_exc(file=handle)

                        self.observer.warn(
                            "unhandled exception caught and "
                            f"suppressed:\n{handle.getvalue()}"
                        )
                finally:
                    self.observer.trigger_end(hook, trigger)
        finally:
            self.phase = None

    @staticmethod
    def generate_offset_cset(engine, csets, cset_generator):
        """Generate a cset with offset applied."""
        return cset_generator(engine, csets).insert_offset(engine.offset)

    @staticmethod
    def get_pkg_contents(engine, csets, pkg):
        """Generate the cset of what files shall be merged to the livefs."""
        return pkg.contents.clone()

    @staticmethod
    def get_remove_cset(engine, csets):
        """Generate the cset of what files shall be removed from the livefs."""
        return csets["old_cset"].difference(csets["install"])

    @staticmethod
    def get_replace_cset(engine, csets):
        """Return the cset of what will be replaced going from old->new pkg."""
        return csets["install"].intersection(csets["old_cset"])

    @staticmethod
    def _get_livefs_intersect_cset(engine, csets, cset_name, realpath=False):
        """Generate the livefs intersection against a cset."""
        return contents.contentsSet(livefs.intersect(csets[cset_name], realpath=realpath))

    @staticmethod
    def get_install_livefs_intersect(engine, csets):
        return engine._get_livefs_intersect_cset(engine, csets, "install")

    @staticmethod
    def get_uninstall_livefs_intersect(engine, csets):
        return engine._get_livefs_intersect_cset(engine, csets, "raw_old_cset")

    alias_cset = staticmethod(alias_cset)

    def get_merged_cset(self, strip_offset=True):
        cset = self.csets["install"]
        if self.offset not in (None, '/') and strip_offset:
            rewrite = contents.change_offset_rewriter(self.offset, '/', cset)
            cset = contents.contentsSet(rewrite)
        return cset

    def get_writable_fsobj(self, fsobj, prefer_reuse=True, empty=False):
        path = source = None
        if fsobj:
            source = fsobj.data
            if source.mutable:
                return fsobj
            if self.allow_reuse and prefer_reuse:
                path = source.path

                # XXX: this should be doing abspath fs intersection probably,
                # although the paths generated are from triggers/engine- still.

                if path is not None and not path.startswith(self.tempdir):
                    # the fsobj pathway isn't in temp space; force a transfer.
                    path = None

            if path:
                # ok, it's tempspace, and reusable.
                obj = data_source.local_source(path, True, encoding=source.encoding)

                if empty:
                    obj.bytes_fileobj(True).truncate(0)
                return obj

        # clone it into tempspace; it's required we control the tempspace,
        # so this function is safe in our usage.
        fd, path = tempfile.mkstemp(prefix='merge-engine-', dir=self.tempdir)

        # XXX: annoying quirk of python, we don't want append mode, so 'a+'
        # isn't viable; wr will truncate the file, so data_source uses r+.
        # this however doesn't allow us to state "create if missing"
        # so we create it ourselves. Annoying, but so it goes.
        # just touch the filepath.
        touch(path)
        new_source = data_source.local_source(
            path, True, encoding=getattr(fsobj, 'encoding', None))

        if source and not empty:
            data_source.transfer(source.bytes_fsobj(), new_source.bytes_fsobj(True))
        return new_source



"""
built ebuild packages (vdb packages and binpkgs are derivatives of this)
"""

__all__ = ("package", "package_factory")

import itertools
from functools import partial

from snakeoil.currying import post_curry
from snakeoil.data_source import local_source
from snakeoil.mappings import IndeterminantDict
from snakeoil.obj import DelayedInstantiation

from .. import fetch
from ..fs.livefs import scan
from ..merge import engine
from ..package import metadata
from ..package.base import DynamicGetattrSetter
from . import conditionals, ebd, ebuild_src, triggers
from .eapi import get_eapi


def _passthrough(inst, attr):
    return inst.data[attr]


def _chost_fallback(initial, self):
    o = self.data.get(initial)
    if o is None:
        o = self.data.get("CHOST")
        if o is None:
            return o
    return o.strip()


def _render_and_evaluate_attr(self, attr_func, render_func):
    return render_func(attr_func(self), self.use)


class package(ebuild_src.base):
    """Built form of an ebuild."""

    immutable = True
    allow_regen = False

    __slots__ = ()

    # Data in a 'built' ebuild may not be stored in finalized/rendered form, thus
    # for all configurables, for it to be rendered if accessed.
    # Note that since _get_attr doesn't  vary across EAPI, we can ignore tracked_attributes-
    # we render either way (because the underlying API irregardless of EAPI must be usable, even
    # if returned data is effectively empty).  Finally, note that this just maps the list across;
    # it's expected that certain attributes that are known to have no meaning for a 'built' package
    # are nulled (for example, fetchables: nothing to fetch).
    locals().update({
        attr_name: DynamicGetattrSetter.register(
            post_curry(
                _render_and_evaluate_attr,
                ebuild_src.package._get_attr[attr_name],
                render_func
            )
        )
        for attr_name, render_func in ebuild_src.package._config_wrappables.items()
    })

    @property
    def _operations(self):
        return ebd.built_operations

    # hack, not for consumer usage
    _is_from_source = False

    built = True

    cbuild = DynamicGetattrSetter.register(partial(_chost_fallback, 'CBUILD'))
    chost = DynamicGetattrSetter.register(partial(_chost_fallback, 'CHOST'))
    ctarget = DynamicGetattrSetter.register(partial(_chost_fallback, 'CTARGET'))
    contents = DynamicGetattrSetter.register(post_curry(_passthrough, 'contents'))
    environment = DynamicGetattrSetter.register(post_curry(_passthrough, 'environment'))

    @property
    def tracked_attributes(self):
        # tracked attributes varies depending on EAPI, thus this has to be runtime computed
        return tuple(itertools.chain(
            super().tracked_attributes, ('contents', 'use', 'environment')
        ))

    @DynamicGetattrSetter.register
    def cflags(self):
        return self.data.get("CFLAGS", "")

    @DynamicGetattrSetter.register
    def cxxflags(self):
        return self.data.get("CXXFLAGS", "")

    @DynamicGetattrSetter.register
    def ldflags(self):
        return self.data.get("LDFLAGS", "")

    @DynamicGetattrSetter.register
    def distfiles(self):
        return tuple(self.data.get("DISTFILES", "").split())

    @DynamicGetattrSetter.register
    def iuse_effective(self):
        return tuple(self.data.get("IUSE_EFFECTIVE", "").split())

    @DynamicGetattrSetter.register
    def inherited(self):
        return tuple(self.data.get("INHERITED", "").split())

    @DynamicGetattrSetter.register
    def inherit(self):
        return tuple(self.data.get("INHERIT", "").split())

    @DynamicGetattrSetter.register
    def source_repository(self):
        repo = self.data.get('source_repository')
        if repo is None:
            repo = self.data.get('repository')
            # work around managers storing this in different places.
            if repo is None:
                # finally, do the strip ourselves since this can come
                # back as '\n' from binpkg Packages caches...
                repo = self.data.get('REPO', '').strip()
                if not repo:
                    repo = None
        if isinstance(repo, str):
            repo = repo.strip()
        return repo if repo else None

    @DynamicGetattrSetter.register
    def fetchables(self, ret=conditionals.DepSet.parse('', fetch.fetchable, operators={})):
        return ret

    @DynamicGetattrSetter.register
    def use(self):
        return DelayedInstantiation(
            frozenset, lambda: frozenset(self.data["USE"].split())
        )

    @DynamicGetattrSetter.register
    def eapi(self):
        eapi_magic = self.data.pop("EAPI", "0")
        if not eapi_magic:
            # "" means EAPI 0
            eapi_magic = '0'
        eapi = get_eapi(str(eapi_magic).strip())
        # This can return None... definitely the wrong thing right now
        # for an unsupported eapi. Fix it later.
        return eapi

    def _update_metadata(self, pkg):
        raise NotImplementedError()

    def _repo_install_op(self, domain, observer):
        return self._parent._generate_format_install_op(domain, self, observer)

    def _repo_uninstall_op(self, domain, observer):
        return self._parent._generate_format_uninstall_op(domain, self, observer)

    def _repo_replace_op(self, domain, old_pkg, observer):
        return self._parent._generate_format_replace_op(domain, old_pkg, self, observer)

    def _fetch_metadata(self):
        return self._parent._get_metadata(self)

    def __str__(self):
        return "built ebuild: %s" % (self.cpvstr)

    def build(self, **kwargs):
        return self.repo._generate_buildop(self)

    def add_format_triggers(self, *args, **kwds):
        return self._parent._add_format_triggers(self, *args, **kwds)

    @property
    def ebuild(self):
        o = self.data.get("ebuild")
        if o is not None:
            return o
        return self._parent._get_ebuild_src(self)

    @property
    def _mtime_(self):
        raise AttributeError(self, "_mtime_")


class fresh_built_package(package):

    __slots__ = ()
    _is_from_source = True


def generic_format_triggers(self, pkg, op_inst, format_op_inst, engine_inst):
    if (engine_inst.mode in (engine.REPLACE_MODE, engine.INSTALL_MODE)
            and pkg == engine_inst.new and pkg.repo is engine_inst.new.repo):
        if 'preinst' in pkg.mandatory_phases:
            t = triggers.preinst_contents_reset(format_op_inst)
            t.register(engine_inst)
        # for ebuild format, always check the syms.
        # this isn't perfect for binpkgs since if the binpkg is already
        # screwed, the target is in place already
        triggers.FixImageSymlinks(format_op_inst).register(engine_inst)


def _generic_format_install_op(self, domain, newpkg, observer):
    return ebd.install_op(domain, newpkg, observer)


def _generic_format_uninstall_op(self, domain, oldpkg, observer):
    return ebd.uninstall_op(domain, oldpkg, observer)


def _generic_format_replace_op(self, domain, oldpkg, newpkg, observer):
    return ebd.replace_op(domain, oldpkg, newpkg, observer)


class package_factory(metadata.factory):
    child_class = package

    # For the plugin system.
    priority = 5

    def _get_metadata(self, pkg):
        return self._parent_repo._get_metadata(pkg)

    def new_package(self, *args):
        inst = self._cached_instances.get(args)
        if inst is None:
            inst = self._cached_instances[args] = self.child_class(self, *args)
        return inst

    def _get_ebuild_path(self, pkg):
        return self._parent_repo._get_path(pkg)

    _generate_format_install_op = _generic_format_install_op
    _generate_format_uninstall_op = _generic_format_uninstall_op
    _generate_format_replace_op = _generic_format_replace_op
    _add_format_triggers = generic_format_triggers


class fake_package_factory(package_factory):
    """A fake package_factory, so that we can reuse the normal get_metadata hooks.

    A factory is generated per package instance, rather then one
    factory, N packages.

    Do not use this unless you know it's what your after; this is
    strictly for transitioning a built ebuild (still in the builddir)
    over to an actual repo. It literally is a mapping of original
    package data to the new generated instances data store.
    """

    def __init__(self, child_class):
        self.child_class = child_class
        self._parent_repo = None

    def new_package(self, pkg, image_root, environment_path):
        self.pkg = pkg
        self.image_root = image_root
        self.environment_path = environment_path
        # lambda redirects path to environment path
        obj = self.child_class(self, pkg.cpvstr)
        for x in self.pkg._raw_pkg.tracked_attributes:
            # bypass setattr restrictions.
            object.__setattr__(obj, x, getattr(self.pkg, x))
        object.__setattr__(obj, "use", self.pkg.use)
        object.__setattr__(obj, "_domain", self.pkg._domain)
        return obj

    def get_ebuild_src(self, pkg):
        return self.pkg.ebuild

    def scan_contents(self, location):
        return scan(location, offset=location, mutable=True)

    def _get_metadata(self, pkg):
        return IndeterminantDict(self.__pull_metadata)

    def __pull_metadata(self, key):
        if key == "contents":
            return self.scan_contents(self.image_root)
        elif key == "environment":
            return local_source(self.environment_path)
        else:
            try:
                return getattr(self.pkg, key)
            except AttributeError as e:
                raise KeyError(key) from e

    _generate_format_install_op = _generic_format_install_op
    _generate_format_uninstall_op = _generic_format_uninstall_op
    _generate_format_replace_op = _generic_format_replace_op
    _add_format_triggers = generic_format_triggers


generate_new_factory = package_factory



"""
filtering repository
"""

__all__ = ("tree",)

from itertools import filterfalse

from snakeoil.klass import DirProxy, GetAttrProxy

from ..operations.repo import operations_proxy
from ..restrictions import restriction
from . import errors, prototype


class tree(prototype.tree):
    """Filter existing repository based upon passed in restrictions."""

    operations_kls = operations_proxy

    def __init__(self, repo, restrict, sentinel_val=False):
        self.raw_repo = repo
        self.sentinel_val = sentinel_val
        if not hasattr(self.raw_repo, 'itermatch'):
            raise errors.InitializationError(
                f"{self.raw_repo} is not a repository tree derivative")
        if not isinstance(restrict, restriction.base):
            raise errors.InitializationError(f"{restrict} is not a restriction")
        self.restrict = restrict
        self.raw_repo = repo
        if sentinel_val:
            self._filterfunc = filter
        else:
            self._filterfunc = filterfalse

    def itermatch(self, restrict, **kwds):
        # note that this lets the repo do the initial filtering.
        # better design would to analyze the restrictions, and inspect
        # the repo, determine what can be done without cost
        # (determined by repo's attributes) versus what does cost
        # (metadata pull for example).
        return self._filterfunc(
            self.restrict.match, self.raw_repo.itermatch(restrict, **kwds))

    itermatch.__doc__ = prototype.tree.itermatch.__doc__.replace(
        "@param", "@keyword").replace(":keyword restrict:", ":param restrict:")

    def __len__(self):
        count = 0
        for i in self:
            count += 1
        return count

    __getattr__ = GetAttrProxy("raw_repo")
    __dir__ = DirProxy("raw_repo")

    def __getitem__(self, key):
        v = self.raw_repo[key]
        if self.restrict.match(v) != self.sentinel_val:
            raise KeyError(key)
        return v

    def __repr__(self):
        return '<%s raw_repo=%r restrict=%r sentinel=%r @%#8x>' % (
            self.__class__.__name__,
            getattr(self, 'raw_repo', 'unset'),
            getattr(self, 'restrict', 'unset'),
            getattr(self, 'sentinel_val', 'unset'),
            id(self))



__all__ = ("http_syncer",)

import errno
import os
import ssl
import sys
import urllib.request

from snakeoil.fileutils import AtomicWriteFile, readfile_ascii
from snakeoil.osutils import pjoin

from ..log import logger
from . import base


class http_syncer(base.Syncer):
    """Syncer that fetches files over HTTP(S)."""

    forcable = True

    def __init__(self, basedir, uri, dest=None, **kwargs):
        self.basename = os.path.basename(uri)
        super().__init__(basedir, uri, **kwargs)

    def _sync(self, verbosity, output_fd, force=False, **kwargs):
        dest = self._pre_download()

        if self.uri.lower().startswith('https://'):
            # default to using system ssl certs
            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        else:
            context = None

        headers = {}
        etag_path = pjoin(self.basedir, '.etag')
        modified_path = pjoin(self.basedir, '.modified')

        if not force:
            # use cached ETag to check if updates exist
            previous_etag = readfile_ascii(etag_path, none_on_missing=True)
            if previous_etag:
                headers['If-None-Match'] = previous_etag

            # use cached modification timestamp to check if updates exist
            previous_modified = readfile_ascii(modified_path, none_on_missing=True)
            if previous_modified:
                headers['If-Modified-Since'] = previous_modified

        req = urllib.request.Request(self.uri, headers=headers, method='GET')

        # TODO: add customizable timeout
        try:
            resp = urllib.request.urlopen(req, context=context)
        except urllib.error.URLError as e:
            if e.getcode() == 304: # Not Modified
                logger.debug("content is unchanged")
                return True
            raise base.SyncError(f'failed fetching {self.uri!r}: {e.reason}') from e

        # Manually check cached values ourselves since some servers appear to
        # ignore If-None-Match or If-Modified-Since headers.
        convert = lambda x: x.strip() if x else None
        etag = resp.getheader('ETag')
        modified = resp.getheader('Last-Modified')
        if not force:
            if etag is not None and convert(etag) == convert(previous_etag):
                logger.debug(f"etag {etag} is equal, no update available")
                return True
            if modified is not None and convert(modified) == convert(previous_modified):
                logger.debug(f"header mtime is unmodified: {modified}")
                return True

        try:
            os.makedirs(self.basedir, exist_ok=True)
        except OSError as e:
            raise base.SyncError(
                f'failed creating repo dir {self.basedir!r}: {e.strerror}') from e

        length = resp.getheader('content-length')
        if length:
            length = int(length)
            blocksize = max(4096, length // 100)
        else:
            blocksize = 1000000

        try:
            self._download = AtomicWriteFile(dest, binary=True, perms=0o644)
        except OSError as e:
            raise base.PathError(self.basedir, e.strerror) from e

        # retrieve the file while providing simple progress output
        size = 0
        while True:
            buf = resp.read(blocksize)
            if not buf:
                if length:
                    sys.stdout.write('\n')
                break
            self._download.write(buf)
            size += len(buf)
            if length:
                sys.stdout.write('\r')
                progress = '=' * int(size / length * 50)
                percent = int(size / length * 100)
                sys.stdout.write("[%-50s] %d%%" % (progress, percent))
                sys.stdout.flush()

        self._post_download(dest)

        # TODO: store this in pkgcore cache dir instead?
        # update cached ETag/Last-Modified values
        if etag:
            with open(etag_path, 'w') as f:
                f.write(etag)
        if modified:
            with open(modified_path, 'w') as f:
                f.write(modified)

        return True

    def _pre_download(self):
        """Pre-download initialization.

        Returns file path to download file to.
        """
        return pjoin(self.basedir, self.basename)

    def _post_download(self, path):
        """Post-download file processing.

        Args:
            path (str): path to downloaded file
        """
        # atomically create file
        self._download.close()



"""plugin cache update utility"""

from functools import partial

from snakeoil.sequences import stable_unique

from .. import plugin, plugins
from ..util import commandline

argparser = commandline.ArgumentParser(
    config=False, domain=False, color=False,
    description=__doc__, script=(__file__, __name__))
argparser.add_argument(
    "packages", nargs="*", action='store', default=[plugins],
    type=partial(commandline.python_namespace_type, module=True),
    help="python namespace(s) to regenerate plugins for.  If none are "
         "specified, pkgcore.plugins is updated")


@argparser.bind_main_func
def main(options, out, err):
    """Update caches."""
    for package in stable_unique(options.packages):
        if options.verbosity >= 0:
            out.write(f'updating cache: {package.__name__}')
        plugin.initialize_cache(package, force=True)



"""
remote binpkg support

Currently this primarily just holds the Packages cache used for remote, and
local binpkg repositories
"""

__all__ = ("PackagesCacheV0", "PackagesCacheV1")

import os
from operator import itemgetter
from time import time

from snakeoil.chksum import get_chksums
from snakeoil.containers import RefCountingSet
from snakeoil.fileutils import AtomicWriteFile, readlines
from snakeoil.mappings import ImmutableDict, StackedDict

from .. import cache
from ..log import logger
from ..restrictions import packages


def _iter_till_empty_newline(data):
    for x in data:
        if not x:
            return
        k, v = x.split(':', 1)
        yield k, v.strip()


class CacheEntry(StackedDict):
    """Customized version of StackedDict blocking pop from modifying the target.

    Note that this pop doesn't through KeyError if something is missing- just
    returns None instead. This is likely to be changed.
    """
    def pop(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default


def find_best_savings(stream, line_prefix):
    rcs = RefCountingSet(stream)
    line_overhead = len(line_prefix)
    stream = ((k, v) for k, v in rcs.items() if v != 1)
    return max(stream, key=lambda x: (len(x[0]) + line_overhead) * x[1])[0]


class PackagesCacheV0(cache.bulk):
    """Cache backend for writing binpkg Packages caches

    Note this follows version 0 semantics- not the most efficient, and
    doesn't bundle certain useful keys like RESTRICT
    """

    _header_mangling_map = ImmutableDict({
        'FEATURES': 'UPSTREAM_FEATURES',
        'ACCEPT_KEYWORDS': 'KEYWORDS',
    })

    # this maps from literal keys in the cache to .data[key] expected forms
    _deserialize_map = {
        'DESC': 'DESCRIPTION',
        'MTIME': 'mtime',
        'repo': 'REPO',
    }
    # this maps from .attr to data items.
    _serialize_map = {
        'DESCRIPTION': 'DESC',
        'mtime': 'MTIME',
        'source_repository': 'REPO',
    }
    deserialized_inheritable = frozenset(('CBUILD', 'CHOST', 'source_repository'))
    _pkg_attr_sequences = ('use', 'keywords', 'iuse')
    _deserialized_defaults = dict.fromkeys(
        (
            'BDEPEND', 'DEPEND', 'RDEPEND', 'PDEPEND', 'IDEPEND',
            'BUILD_TIME', 'IUSE', 'KEYWORDS', 'LICENSE', 'PATH', 'PROPERTIES',
            'USE', 'DEFINED_PHASES', 'CHOST', 'CBUILD', 'DESC', 'REPO',
            'DESCRIPTION',
        ),
        ''
    )
    _deserialized_defaults.update({'EAPI': '0', 'SLOT': '0'})
    _deserialized_defaults = ImmutableDict(_deserialized_defaults)

    _stored_chfs = ('size', 'sha1', 'md5', 'mtime')

    version = 0

    def __init__(self, location, *args, **kwds):
        self._location = location
        vkeys = {'CPV'}
        vkeys.update(self._deserialized_defaults)
        vkeys.update(x.upper() for x in self._stored_chfs)
        kwds["auxdbkeys"] = vkeys
        super().__init__(*args, **kwds)

    def _handle(self):
        return readlines(self._location, True, False, False)

    def read_preamble(self, handle):
        return ImmutableDict(
            (self._header_mangling_map.get(k, k), v)
            for k, v in _iter_till_empty_newline(handle))

    def _read_data(self):
        try:
            handle = self._handle()
        except FileNotFoundError:
            return {}
        self.preamble = self.read_preamble(handle)

        defaults = dict(self._deserialized_defaults.items())
        defaults.update((k, v) for k, v in self.preamble.items()
                        if k in self.deserialized_inheritable)
        defaults = ImmutableDict(defaults)

        pkgs = {}
        count = 0
        vkeys = self._known_keys
        while True:
            raw_d = dict(_iter_till_empty_newline(handle))

            d = {k: v for k, v in raw_d.items() if k in vkeys}
            if not d:
                break
            count += 1
            cpv = d.pop("CPV", None)
            if cpv is None:
                cpv = f"{d.pop('CATEGORY')}/{d.pop('PF')}"

            if 'USE' in d:
                d.setdefault('IUSE', d.get('USE', ''))
            for src, dst in self._deserialize_map.items():
                if src in d:
                    d.setdefault(dst, d.pop(src))

            pkgs[cpv] = CacheEntry(d, defaults)
        assert count == int(self.preamble.get('PACKAGES', count))
        return pkgs

    @classmethod
    def _assemble_preamble_dict(cls, target_dicts):
        preamble = {
            'VERSION': cls.version,
            'PACKAGES': len(target_dicts),
            'TIMESTAMP': str(int(time())),
        }
        for key in cls.deserialized_inheritable:
            try:
                preamble[key] = find_best_savings(
                    (d[1].get(key, '') for d in target_dicts), key)
            except ValueError:
                # empty iterable handed to max
                pass
        return preamble

    @classmethod
    def _assemble_pkg_dict(cls, pkg):
        d = {}
        sequences = cls._pkg_attr_sequences
        for key in cls._stored_attrs:

            value = getattr(pkg, key)
            if key in sequences:
                value = ' '.join(sorted(value))
            else:
                value = str(getattr(pkg, key)).strip()
            key = key.upper()
            d[cls._serialize_map.get(key, key)] = value

        for key, value in zip(cls._stored_chfs,
                               get_chksums(pkg.path, *cls._stored_chfs)):
            if key != 'size':
                value = "%x" % (value,)
            d[key.upper()] = value
        d["MTIME"] = str(os.stat(pkg.path).st_mtime)
        return d

    def _write_data(self):
        handler = None
        try:
            try:
                handler = AtomicWriteFile(self._location)
                self._serialize_to_handle(list(self.data.items()), handler)
                handler.close()
            except PermissionError as e:
                logger.error(
                    f'failed writing binpkg cache to {self._location!r}: {e}')
        finally:
            if handler is not None:
                handler.discard()

    def _serialize_to_handle(self, data, handler):
        preamble = self._assemble_preamble_dict(data)

        convert_key = self._serialize_map.get

        for key in sorted(preamble):
            handler.write(f"{convert_key(key, key)}: {preamble[key]}\n")
        handler.write('\n')

        spacer = ' '
        if self.version != 0:
            spacer = ''

        vkeys = self._known_keys
        for cpv, pkg_data in sorted(data, key=itemgetter(0)):
            handler.write(f"CPV:{spacer}{cpv}\n")
            data = [(convert_key(key, key), value)
                    for key, value in pkg_data.items()]
            for write_key, value in sorted(data):
                if write_key not in vkeys:
                    continue
                value = str(value).strip()
                if write_key in preamble:
                    if value != preamble[write_key]:
                        if value:
                            handler.write(f"{write_key}:{spacer}{value}\n")
                        else:
                            handler.write(f"{write_key}:\n")
                elif value:
                    handler.write(f"{write_key}:{spacer}{value}\n")
            handler.write('\n')

    def update_from_xpak(self, pkg, xpak):
        # invert the lookups here; if you do .items() on an xpak,
        # it'll load up the contents in full.
        new_dict = {k: xpak[k] for k in self._known_keys if k in xpak}
        new_dict['_chf_'] = xpak._chf_
        chfs = [x for x in self._stored_chfs if x != 'mtime']
        for key, value in zip(chfs, get_chksums(pkg.path, *chfs)):
            if key != 'size':
                value = "%x" % (value,)
            new_dict[key.upper()] = value
        self[pkg.cpvstr] = new_dict
        return new_dict

    def update_from_repo(self, repo):
        # try to collapse certain keys down to the profile preamble
        targets = repo.match(packages.AlwaysTrue, sorter=sorted)

        if not targets:
            # just open/trunc the target instead, and bail
            open(self._location, 'wb').close()
            return


class PackagesCacheV1(PackagesCacheV0):
    """Cache backend for writing binpkg Packages caches in format version 1.

    See :py:class:`PackagesCacheV0` for usage information; this just writes
    a better ondisk format.
    """

    deserialized_inheritable = PackagesCacheV0.deserialized_inheritable.union(
        ('SLOT', 'EAPI', 'LICENSE', 'KEYWORDS', 'USE', 'RESTRICT'))

    _deserialized_defaults = ImmutableDict(
        list(PackagesCacheV0._deserialized_defaults.items()) + [('RESTRICT', '')])

    @classmethod
    def _assemble_pkg_dict(cls, pkg):
        # not the most efficient...
        d = PackagesCacheV0._assemble_pkg_dict(pkg)
        use = set(pkg.use).intersection(pkg.iuse_stripped)
        d.pop("IUSE", None)
        iuse_bits = [f'-{x}' for x in pkg.iuse_stripped if x not in use]
        use.update(iuse_bits)
        d["USE"] = ' '.join(sorted(use))
        return d

    version = 1


def get_cache_kls(version):
    version = str(version)
    if version == '0':
        return PackagesCacheV0
    elif version in ('1', '-1'):
        return PackagesCacheV1
    raise KeyError(f"cache version {version} unsupported")



from functools import partial

from pkgcore.package.base import base, dynamic_getattr_dict
from pkgcore.package.mutated import MutatedPkg
from snakeoil.compatibility import cmp
from snakeoil.klass import inject_richcmp_methods_from_cmp
from snakeoil.test import TestCase


def passthru(val, self):
    return val

class FakePkg(base):

    # XXX why isn't this using existing classes?
    __slotting_intentionally_disabled__ = True

    def __init__(self, pkg, ver, data):
        base.__init__(self)
        self.pkg = pkg
        self.ver = ver
        self._get_attr = {k: partial(passthru, v) for k, v in data.items()}

    # disable protection.  don't want it here
    __setattr__ = object.__setattr__
    __getattr__ = dynamic_getattr_dict

    def __cmp__(self, other):
        return cmp(self.ver, other.ver)

    inject_richcmp_methods_from_cmp(locals())


class TestMutatedPkg(TestCase):

    def make_fakepkg(self, pkg="dar", ver=1, data=None):
        if data is None:
            data = {"a":1}
        return FakePkg(pkg, ver, data)

    def test_raw_pkg(self):
        pkg = self.make_fakepkg()
        self.assertIdentical(MutatedPkg(pkg, {})._raw_pkg, pkg)

    def test_cmp(self):
        pkg1 = self.make_fakepkg()
        pkg2 = self.make_fakepkg(ver=2)
        mpkg1 = MutatedPkg(pkg1, {})
        mpkg2 = MutatedPkg(pkg2, {})

        for lpkg in (pkg1, mpkg1):
            self.assertTrue(cmp(lpkg, mpkg2) < 0)
            self.assertTrue(cmp(mpkg2, lpkg) > 0)
        self.assertEqual(mpkg1, mpkg1)
        self.assertEqual(pkg1, mpkg1)

    def test_getattr(self):
        pkg = self.make_fakepkg()
        self.assertEqual(MutatedPkg(pkg, {}).a, 1)
        self.assertEqual(MutatedPkg(pkg, {"a":2}).a, 2)
        self.assertRaises(AttributeError, MutatedPkg(pkg, {}).__getattr__, "b")



from snakeoil.test.slot_shadowing import SlotShadowing


class Test_slot_shadowing(SlotShadowing):

    target_namespace = "pkgcore"
    ignore_all_import_failures = True

    def _default_module_blacklister(self, target):
        return target in self.module_blacklist or target.startswith("pkgcore.test.")



from snakeoil.test.mixins import PythonNamespaceWalker


class walker(PythonNamespaceWalker):

    ignore_all_import_failures = True

    def _default_module_blacklister(self, target):
        if target.startswith("pkgcore.test.") or target.startswith('pkgcore.plugins.') \
                or 'pkgcore.test' == target:
            return True
        return PythonNamespaceWalker._default_module_blacklister(self, target)

targets = []
for module in walker().walk_namespace('pkgcore'):
    for name in dir(module):
        obj = getattr(module, name)
        if getattr(obj, 'pkgcore_config_type', None) is not None:
            targets.append(f'{module.__name__}.{name}')

pkgcore_plugins = {
    'configurable': targets
}



"""
fetcher class that pulls files via executing another program to do the fetching
"""

__all__ = ("MalformedCommand", "fetcher",)

import os

from snakeoil.osutils import pjoin
from snakeoil.process.spawn import is_userpriv_capable, spawn_bash

from ..config.hint import ConfigHint
from ..os_data import portage_gid, portage_uid
from . import base, errors, fetchable


class MalformedCommand(errors.FetchError):

    def __init__(self, command):
        super().__init__(f'fetchcommand is malformed: {command}')
        self.command = command


class fetcher(base.fetcher):

    pkgcore_config_type = ConfigHint(
        {'userpriv': 'bool', 'required_chksums': 'list',
         'distdir': 'str', 'command': 'str', 'resume_command': 'str'},
        allow_unknowns=True)

    def __init__(self, distdir, command, resume_command=None,
                 required_chksums=None, userpriv=True, attempts=10,
                 readonly=False, **extra_env):
        """
        :param distdir: directory to download files to
        :type distdir: string
        :param command: shell command to execute to fetch a file
        :type command: string
        :param resume_command: if not None, command to use for resuming-
            if None, command is reused
        :param required_chksums: if None, all chksums must be verified,
            else only chksums listed
        :type required_chksums: None or sequence
        :param userpriv: depriv for fetching?
        :param attempts: max number of attempts before failing the fetch
        :param readonly: controls whether fetching is allowed
        """
        super().__init__()
        self.distdir = distdir
        if required_chksums is not None:
            required_chksums = [x.lower() for x in required_chksums]
        else:
            required_chksums = []
        if len(required_chksums) == 1 and required_chksums[0] == "all":
            self.required_chksums = None
        else:
            self.required_chksums = required_chksums

        def rewrite_command(string):
            new_command = string.replace("${DISTDIR}", self.distdir)
            new_command = new_command.replace("$DISTDIR", self.distdir)
            new_command = new_command.replace("${URI}", "%(URI)s")
            new_command = new_command.replace("$URI", "%(URI)s")
            new_command = new_command.replace("${FILE}", "%(FILE)s")
            new_command = new_command.replace("$FILE", "%(FILE)s")
            if new_command == string:
                raise MalformedCommand(string)
            try:
                new_command % {"URI": "blah", "FILE": "blah"}
            except KeyError as k:
                raise MalformedCommand(f"{command}: unexpected key {k.args[0]}")
            return new_command

        self.command = rewrite_command(command)
        if resume_command is None:
            self.resume_command = self.command
        else:
            self.resume_command = rewrite_command(resume_command)

        self.attempts = attempts
        self.userpriv = userpriv
        self.readonly = readonly
        self.extra_env = extra_env

    def fetch(self, target):
        """Fetch a file.

        :type target: :obj:`pkgcore.fetch.fetchable` instance
        :return: None if fetching failed,
            else on disk location of the copied file
        """
        if not isinstance(target, fetchable):
            raise TypeError(
                f"target must be fetchable instance/derivative: {target}")

        path = pjoin(self.distdir, target.filename)
        uris = iter(target.uri)
        last_exc = RuntimeError("fetching failed for an unknown reason")
        spawn_opts = {'umask': 0o002, 'env': self.extra_env}
        if self.userpriv and is_userpriv_capable():
            spawn_opts.update({"uid": portage_uid, "gid": portage_gid})

        for _attempt in range(self.attempts):
            try:
                self._verify(path, target)
                return path
            except errors.MissingDistfile as e:
                command = self.command
                last_exc = e
            except errors.ChksumFailure:
                raise
            except errors.FetchFailed as e:
                last_exc = e
                if not e.resumable:
                    try:
                        os.unlink(path)
                        command = self.command
                    except OSError as e:
                        raise errors.UnmodifiableFile(path, e) from e
                else:
                    command = self.resume_command
            # Note we're not even checking the results, the verify portion of
            # the loop handles this. In other words, don't trust the external
            # fetcher's exit code, trust our chksums instead.
            try:
                spawn_bash(
                    command % {"URI": next(uris), "FILE": target.filename},
                    **spawn_opts)
            except StopIteration:
                raise errors.FetchFailed(
                    target.filename, "ran out of urls to fetch from")
        else:
            raise last_exc

    def get_path(self, fetchable):
        path = pjoin(self.distdir, fetchable.filename)
        if self._verify(path, fetchable) is None:
            return path
        return None



from pkgcore import fetch
from snakeoil.sequences import iflatten_instance
from snakeoil.test import TestCase


class base(TestCase):

    def assertUri(self, obj, uri):
        uri = list(uri)
        self.assertEqual(list(iflatten_instance(obj)), uri)
        if uri:
            self.assertTrue(obj)
        else:
            self.assertFalse(obj)


class TestFetchable(base):

    def test_init(self):
        o = fetch.fetchable("dar", uri=["asdf"], chksums={"asdf":1})
        self.assertEqual(o.filename, "dar")
        self.assertUri(o.uri, ["asdf"])
        self.assertEqual(o.chksums, {"asdf":1})

    def test_eq_ne(self):
        o1 = fetch.fetchable("dar", uri=["asdf"], chksums={"asdf":1})
        self.assertEqual(o1, o1)
        o2 = fetch.fetchable("dar", uri=["asdf"], chksums={"asdf":1})
        self.assertEqual(o1, o2)
        self.assertNotEqual(o1,
            fetch.fetchable("dar1", uri=["asdf"], chksums={"asdf":1}))
        self.assertNotEqual(o1,
            fetch.fetchable("dar", uri=["asdf1"], chksums={"asdf":1}))
        self.assertNotEqual(o1,
            fetch.fetchable("dar", uri=["asdf1"], chksums={"asdf":1, "foon":1}))


class TestMirror(base):

    kls = fetch.mirror

    default_mirrors = ["http://foon", "ftp://spoon"]
    def setUp(self):
        self.mirror = self.kls(self.default_mirrors, "fork")

    def test_init(self):
        self.assertEqual(self.mirror.mirror_name, "fork")
        # explicit test should any tuple like sequence show up
        self.assertInstance(self.mirror.mirrors, tuple)
        self.assertEqual(self.mirror.mirrors, tuple(self.default_mirrors))

    def test_iter(self):
        self.assertEqual(list(self.mirror), self.default_mirrors)

    def test_len(self):
        self.assertEqual(len(self.mirror), len(self.default_mirrors))

    def test_getitem(self):
        self.assertEqual(self.mirror[1], self.default_mirrors[1])

    def test_eq_ne(self):
        self.assertEqual(self.mirror, self.kls(self.default_mirrors, 'fork'))
        self.assertNotEqual(self.mirror,
            self.kls(self.default_mirrors + ['http://fark'], 'fork'))


class TestDefaultMirror(TestMirror):

    kls = fetch.default_mirror


class Test_uri_list(base):

    def setUp(self):
        self.uril = fetch.uri_list("cows")

    @staticmethod
    def mk_uri_list(*iterable, **kwds):
        filename = kwds.get("filename", "asdf")
        obj = fetch.uri_list(filename)
        for x in iterable:
            if isinstance(x, fetch.mirror):
                obj.add_mirror(x)
            else:
                obj.add_uri(x)
        return obj

    def test_mirrors(self):
        self.assertRaises(TypeError, self.uril.add_mirror, "cows")
        mirror = fetch.mirror(["me", "WI"], "asdf")
        self.uril.add_mirror(mirror)
        self.assertEqual(list(self.uril), ["me/cows", "WI/cows"])
        self.uril.add_mirror(mirror, "foon/boon")
        self.assertUri(self.uril,
            ["me/cows", "WI/cows", "me/foon/boon", "WI/foon/boon"])

    def test_uris(self):
        self.uril.add_uri("blar")
        self.assertUri(self.uril, ["blar"])

    def test_combined(self):
        l = ["blarn", "me/cows", "WI/cows", "madison",
            "belleville/cows", "verona/cows"]
        self.uril.add_uri("blarn")
        self.uril.add_mirror(fetch.mirror(["me", "WI"], "asdf"))
        self.uril.add_uri("madison")
        self.uril.add_mirror(fetch.default_mirror(
            ["belleville", "verona"], "foon"))
        self.assertUri(self.uril, l)

    def test_nonzero(self):
        self.assertTrue(self.mk_uri_list("asdf"))
        self.assertFalse(self.mk_uri_list())
        self.assertFalse(self.mk_uri_list(fetch.mirror((), "mirror")))

    def test_len(self):
        self.assertLen(self.mk_uri_list(), 0)
        self.assertLen(self.mk_uri_list("fdas"), 1)
        self.assertLen(self.mk_uri_list(fetch.mirror((), "mirror")), 0)
        self.assertLen(self.mk_uri_list(fetch.mirror(("asdf",), "mirror")), 1)



import os
import tempfile

from pkgcore import gpg
from pkgcore.ebuild import digest
from snakeoil.data_source import local_source
from snakeoil.test import TestCase

# "Line too long" (and our custom more aggressive version of that)
# pylint: disable-msg=C0301,CPC01

digest_contents = \
"""MD5 98db1465629693fc434d4dc52db93838 Python-2.4.2.tar.bz2 7853169
RMD160 c511d2b76b5394742d285e71570a2bcd3c1fa871 Python-2.4.2.tar.bz2 7853169
SHA256 e163b95ee56819c0f3c58ef9278c30b9e49302c2f1a1917680ca894d33929f7e Python-2.4.2.tar.bz2 7853169
MD5 2fa54dd51b6a8f1c46e5baf741e90f7e python-2.4-patches-1.tar.bz2 7820
RMD160 313c0f4f4dea59290c42a9b2c8de1db159f1ca1b python-2.4-patches-1.tar.bz2 7820
SHA256 e22abe4394f1f0919aac429f155c00ec1b3fe94cdc302119059994d817cd30b5 python-2.4-patches-1.tar.bz2 7820"""
digest_chksum = (
    ("size", int(7853169)),
    ("md5", int("98db1465629693fc434d4dc52db93838", 16)),
    ("rmd160", int("c511d2b76b5394742d285e71570a2bcd3c1fa871", 16)),
    ("sha256", int("e163b95ee56819c0f3c58ef9278c30b9e49302c2f1a1917680ca894d33929f7e", 16))
)

# ripped straight from the glep
pure_manifest2 = \
"""AUX ldif-buffer-overflow-fix.diff 5007 RMD160 1354a6bd2687430b628b78aaf43f5c793d2f0704 SHA1 424e1dfca06488f605b9611160020227ecdd03ac
AUX procmime.patch 977 RMD160 39a51a4d654759b15d1644a79fb6e8921130df3c SHA1 d76929f6dfc2179281f7ccee5789aab4e970ba9e
EBUILD sylpheed-claws-1.0.5-r1.ebuild 3906 RMD160 cdd546c128db2dea7044437de01ec96e12b4f5bf SHA1 a84b49e76961d7a9100852b64c2bfbf9b053d45e
EBUILD sylpheed-claws-1.9.100.ebuild 4444 RMD160 89326038bfc694dafd22f10400a08d3f930fb2bd SHA1 8895342f3f0cc6fcbdd0fdada2ad8e23ce539d23
EBUILD sylpheed-claws-1.9.15.ebuild 4821 RMD160 ec0ff811b893084459fe5b17b8ba8d6b35a55687 SHA1 358278a43da244e1f4803ec4b04d6fa45c41ab4d
MISC ChangeLog 25770 RMD160 0e69dd7425add1560d630dd3367342418e9be776 SHA1 1210160f7baf0319de3b1b58dc80d7680d316d28
MISC metadata.xml 269 RMD160 39d775de55f9963f8946feaf088aa0324770bacb SHA1 4fd7b285049d0e587f89e86becf06c0fd77bae6d
DIST sylpheed-claws-1.0.5.tar.bz2 3268626 RMD160 f2708b5d69bc9a5025812511fde04eca7782e367 SHA1 d351d7043eef7a875df18a8c4b9464be49e2164b
DIST sylpheed-claws-1.9.100.tar.bz2 3480063 RMD160 72fbcbcc05d966f34897efcc1c96377420dc5544 SHA1 47465662b5470af5711493ce4eaad764c5bf02ca
DIST sylpheed-claws-1.9.15.tar.bz2 3481018 RMD160 b01d1af2df55806a8a8275102b10e389e0d98e94 SHA1 a17fc64b8dcc5b56432e5beb5c826913cb3ad79e
"""
pure_manifest2_chksums = {}
for x in pure_manifest2.split("\n"):
    l = x.split()
    if not l:
        continue
    i = iter(l[3:])
    chksum = [("size", int(l[2]))]
    chksum += [(k.lower(), int(v, 16)) for k, v in zip(i, i)]
    chksum = tuple(chksum)
    pure_manifest2_chksums.setdefault(l[0], {})[l[1]] = chksum
    del chksum, l, i


class TestManifest(TestCase):

    convert_source = staticmethod(lambda x:x)

    def get_manifest(self, data):
        fd, fn = tempfile.mkstemp()
        os.write(fd, data.encode())
        os.close(fd)
        try:
            return digest.parse_manifest(self.convert_source(fn))
        finally:
            os.unlink(fn)

    def test_gpg_filtering(self):
        # intentionally stick gpg signing midway through
        data = pure_manifest2.split("\n")
        s = "\n".join(data[0:2])
        s += f"\n{gpg.sig_header}\nasdf\n{gpg.sig_footer}\n"
        s += "\n".join(data[2:])
        # ensure it can parse it
        (dist, aux, ebuild, misc) = self.get_manifest(s)

    def test_manifest2(self):
        (dist, aux, ebuild, misc) = \
            self.get_manifest(pure_manifest2)

        for dtype, d in (("DIST", dist), ("AUX", aux),
            ("EBUILD", ebuild), ("MISC", misc)):
            req_d = pure_manifest2_chksums[dtype]
            self.assertEqual(sorted(req_d), sorted(d))
            for k, v in req_d.items():
                i1 = sorted(v)
                i2 = sorted(d[k].items())
                self.assertEqual(i1, i2, msg="{i1!r} != {i2!r}\nfor {dtype} {k}")


class TestManifestDataSource(TestManifest):
    convert_source = staticmethod(lambda x: local_source(x))



#!/usr/bin/env python3

import os
import sys

from snakeoil.bash import iter_read_bash
from snakeoil.osutils import listdir_files

from pkgcore.config import load_config
from pkgcore.ebuild.atom import atom
# we use a WorldFile since it *currently* forces unversioned atoms.
from pkgcore.pkgsets.filelist import WorldFile


def main(target_repo, seen, moves):
    # could build the atom from categories/packages, but prefer this;
    # simpler.
    new_seen = set(atom("%s/%s" % x) for x in target_repo.versions)

    new_pkgs = new_seen.difference(seen)
    # this is simpler if pkgsets are... actually sets. ;)
    # can't rely on it however since <0.2 lacks it.
    seen_set = set(seen)
    removed = seen_set.difference(new_seen)

    finished_moves = removed.intersection(moves)
    removed.difference_update(moves)
    in_transit = seen_set.intersection(moves)
    in_transit.difference_update(finished_moves)

    d = {}
    for x in in_transit:
        if moves[x] in new_seen:
            d[x] = moves[x]
    in_transit = d

    for l, prefix in ((new_pkgs, "added pkgs"), (removed, "removed pkgs")):
        if l:
            sys.stdout.write("%s:\n  %s\n\n" %
                (prefix, "\n  ".join(str(x) for x in sorted(l))))

    if finished_moves:
        sys.stdout.write("moved pkgs:\n  %s\n\n" %
            "\n  ".join("%s -> %s" % (k, moves[k])
                for k in sorted(finished_moves)))
    if in_transit:
        sys.stdout.write("pkg moves in transit:\n  %s\n\n" %
            "\n  ".join("%s -> %s" % (k, in_transit[k])
                for k in sorted(in_transit)))

    # just flush the seen fully, simplest.
    for x in seen_set:
        seen.remove(x)
    for x in new_seen:
        seen.add(x)
    return True


def apply_updates(moves, atom_set):
    d = {}
    for src, trg in moves.items():
        if src in atom_set:
            d[src] = trg
            atom_set.remove(src)
            atom_set.add(trg)
    return d


def parse_moves(location):
    pjoin = os.path.join

    # schwartzian comparison, convert it into YYYY-QQ
    def get_key(fname):
        return tuple(reversed(fname.split('-')))

    moves = {}
    for update_file in sorted(listdir_files(location), key=get_key):
        for line in iter_read_bash(pjoin(location, update_file)):
            line = line.split()
            if line[0] != 'move':
                continue
            moves[atom(line[1])] = atom(line[2])
    return moves


if __name__ == '__main__':
    args = sys.argv[1:]
    if len(args) not in (2,3) or "--help" in args or "-h" in args:
        sys.stderr.write("need two args; repository to scan, and "
            "file to store the state info in.\nOptional third arg is "
            "a profiles update directory to scan for moves.\n")
        sys.exit(-1)

    conf = load_config()
    try:
        repo = conf.repo[args[0]]
    except KeyError:
        sys.stderr.write("repository %r wasn't found- known repos\n%r\n" %
            (args[0], list(conf.repo.keys())))
        sys.exit(-2)

    if not os.path.exists(args[1]):
        open(args[1], "w")
    filelist = WorldFile(args[1])
    moves = {}
    if len(args) == 3:
        moves = parse_moves(args[2])
    if main(repo, filelist, moves):
        filelist.flush()
        sys.exit(0)
    sys.exit(1)



import argparse
import operator
from collections import defaultdict
from itertools import chain

from snakeoil.cli import arghparse
from snakeoil.osutils import pjoin
from snakeoil.sequences import split_negations

from ..util import commandline
from . import atom, profiles
from .misc import ChunkedDataDict

commands = []
# info: keywords known
# global known flags, etc


class _base(arghparse.ArgparseCommand):

    @staticmethod
    def _validate_args(parser, namespace):
        path = namespace.profile
        if path is None:
            if namespace.repo is not None:
                # default to the repo's main profiles dir
                path = pjoin(namespace.repo.location, 'profiles')
            else:
                # default to the configured system profile if none is selected
                path = namespace.config.get_default("domain").profile.profile
        else:
            if namespace.repo is not None and getattr(namespace.repo, 'location', False):
                if not path.startswith('/'):
                    path = pjoin(namespace.repo.location, 'profiles', path)
        try:
            stack = profiles.ProfileStack(arghparse.existent_path(path))
        except argparse.ArgumentTypeError as e:
            parser.error(e)
        if stack.node.repoconfig is None:
            parser.error(f"invalid profile path: {path!r}")
        namespace.profile = stack

    def bind_to_parser(self, parser):
        arghparse.ArgparseCommand.bind_to_parser(self, parser)
        parser.add_argument('profile', help='path to the profile to inspect')
        name = self.__class__.__name__
        kwds = {(f'_{name}_suppress'): arghparse.DelayedDefault.wipe(('domain'), 50)}
        parser.set_defaults(**kwds)
        parser.bind_final_check(self._validate_args)
        self._subclass_bind(parser)

    def _subclass_bind(self, parser):
        """override to add more command line options"""


_register_command = commandline.register_command(commands)


class parent(_base, metaclass=_register_command):
    """output the linearized tree of inherited parents

    later lines override earlier lines
    """

    def __call__(self, namespace, out, err):
        if namespace.repo is None:
            out.write("\n".join(x.path for x in namespace.profile.stack))
        else:
            repo_dir = pjoin(namespace.repo.location, 'profiles')
            for x in namespace.profile.stack:
                out.write(x.path[len(repo_dir):].lstrip('/'))


class eapi(_base, metaclass=_register_command):
    """output EAPI support required for reading this profile"""

    def __call__(self, namespace, out, err):
        eapis = set(str(x.eapi) for x in namespace.profile.stack)
        out.write("\n".join(sorted(eapis)))


class status(_base, metaclass=_register_command):
    """output profile status"""

    def __call__(self, namespace, out, err):
        profiles_dir = pjoin(namespace.profile.node.repoconfig.location, 'profiles')
        profile_rel_path = namespace.profile.path[len(profiles_dir):].lstrip('/')
        arch_profiles = namespace.profile.node.repoconfig.arch_profiles
        statuses = [(path, status) for path, status in chain.from_iterable(arch_profiles.values())
                    if path.startswith(profile_rel_path)]
        if len(statuses) > 1:
            for path, status in sorted(statuses):
                out.write(f'{path}: {status}')
        elif statuses:
            out.write(statuses[0][1])


class deprecated(_base, metaclass=_register_command):
    """dump deprecation notices, if any"""

    def __call__(self, namespace, out, err):
        for idx, profile in enumerate(x for x in namespace.profile.stack if x.deprecated):
            if idx:
                out.write()
            out.write(out.bold, out.fg("cyan"), profile.path, out.reset, ":")
            data = profile.deprecated
            if data[0]:
                out.write(" ", out.fg("yellow"), "replacement profile", out.reset, f": {data[0]}")
            if data[1]:
                out.write(" ", out.fg("yellow"), "deprecation message", out.reset, ":")
                for line in data[1].split("\n"):
                    out.write(line, prefix='  ')


class provided(_base, metaclass=_register_command):
    """list all package.provided packages

    Note that these are exact versions- if a dep requires a higher version,
    it's not considered satisfied.
    """

    def __call__(self, namespace, out, err):
        targets = defaultdict(list)
        for pkg in namespace.profile.provides_repo:
            targets[pkg.key].append(pkg)

        for pkg_name, pkgs in sorted(targets.items(), key=operator.itemgetter(0)):
            out.write(
                out.fg("cyan"), pkg_name, out.reset, ": ",
                ", ".join(x.fullver for x in sorted(pkgs)))


class system(_base, metaclass=_register_command):
    """output the system package set"""

    def __call__(self, namespace, out, err):
        for pkg in sorted(namespace.profile.system):
            out.write(str(pkg))


class use_expand(_base, metaclass=_register_command):
    """output the USE_EXPAND configuration for this profile

    Outputs two fields of interest; USE_EXPAND (pseudo use groups), and
    USE_EXPAND_HIDDEN which is immutable by user configuration and use deps
    (primarily used for things like setting the kernel or OS type).
    """

    def __call__(self, namespace, out, err):
        out.write(
            "flags: ",
            ', '.join(sorted(namespace.profile.use_expand)))
        out.write(
            "hidden: ",
            ', '.join(sorted(namespace.profile.use_expand_hidden)))


class iuse_effective(_base, metaclass=_register_command):
    """output the IUSE_EFFECTIVE value for this profile"""

    def __call__(self, namespace, out, err):
        if namespace.profile.iuse_effective:
            out.write(' '.join(sorted(namespace.profile.iuse_effective)))


class masks(_base, metaclass=_register_command):
    """inspect package masks"""

    def __call__(self, namespace, out, err):
        for mask in sorted(namespace.profile.masks):
            out.write(str(mask))


class unmasks(_base, metaclass=_register_command):
    """inspect package unmasks"""

    def __call__(self, namespace, out, err):
        for unmask in sorted(namespace.profile.unmasks):
            out.write(str(unmask))


class bashrcs(_base, metaclass=_register_command):
    """inspect bashrcs"""

    def __call__(self, namespace, out, err):
        for bashrc in namespace.profile.bashrcs:
            out.write(bashrc.path)


class keywords(_base, metaclass=_register_command):
    """inspect package.keywords"""

    def __call__(self, namespace, out, err):
        for pkg, keywords in namespace.profile.keywords:
            out.write(f"{pkg}: {' '.join(keywords)}")


class accept_keywords(_base, metaclass=_register_command):
    """inspect package.accept_keywords"""

    def __call__(self, namespace, out, err):
        for pkg, keywords in namespace.profile.accept_keywords:
            out.write(pkg, autoline=False)
            if keywords:
                out.write(f": {' '.join(keywords)}")
            else:
                out.write()


class _use(_base):

    def _output_use(self, neg, pos):
        neg = ('-' + x for x in neg)
        return ' '.join(sorted(chain(neg, pos)))

    def __call__(self, namespace, out, err):
        global_use = []
        pkg_use = {}

        for k, v in namespace.use.render_to_dict().items():
            if isinstance(k, str):
                for pkg, neg, pos in v:
                    if isinstance(pkg, atom.atom):
                        pkg_neg, pkg_pos = pkg_use.setdefault(pkg, (set(), set()))
                        pkg_neg.update(neg)
                        pkg_pos.update(pos)
                        matched = pkg_neg.intersection(pkg_pos)
                        pkg_pos.difference_update(matched)
                        pkg_neg.difference_update(matched)
            else:
                _, neg, pos = v[0]
                global_use = (neg, pos)

        if global_use:
            out.write(f'*/*: {self._output_use(*global_use)}')
        if pkg_use:
            for pkg, (neg, pos) in sorted(pkg_use.items()):
                if neg or pos:
                    out.write(f'{pkg}: {self._output_use(neg, pos)}')


class use(_use, metaclass=_register_command):
    """inspect enabled USE flags

    Including USE, USE_EXPAND, and package.use settings.
    """

    def __call__(self, namespace, out, err):
        u = ChunkedDataDict()
        u.add_bare_global(*split_negations(namespace.profile.use))
        u.merge(namespace.profile.pkg_use)
        namespace.use = u
        super().__call__(namespace, out, err)


class masked_use(_use, metaclass=_register_command):
    """inspect masked use flags"""

    def __call__(self, namespace, out, err):
        namespace.use = namespace.profile.masked_use
        super().__call__(namespace, out, err)


class stable_masked_use(_use, metaclass=_register_command):
    """inspect stable masked use flags"""

    def __call__(self, namespace, out, err):
        namespace.use = namespace.profile.stable_masked_use
        super().__call__(namespace, out, err)


class forced_use(_use, metaclass=_register_command):
    """inspect forced use flags"""

    def __call__(self, namespace, out, err):
        namespace.use = namespace.profile.forced_use
        super().__call__(namespace, out, err)


class stable_forced_use(_use, metaclass=_register_command):
    """inspect stable forced use flags"""

    def __call__(self, namespace, out, err):
        namespace.use = namespace.profile.stable_forced_use
        super().__call__(namespace, out, err)


class defaults(_base, metaclass=_register_command):
    """inspect defined configuration for this profile

    This is data parsed from make.defaults, containing things like
    ACCEPT_KEYWORDS.
    """

    def _subclass_bind(self, parser):
        parser.add_argument(
            "variables", nargs='*',
            help="if not specified, all settings are displayed"
                 ". If given, output is limited to just those settings if "
                 "they exist")

    def __call__(self, namespace, out, err):
        var_filter = namespace.variables
        if var_filter:
            var_filter = set(var_filter).__contains__
        else:
            var_filter = lambda x: True

        settings = namespace.profile.default_env
        vars = sorted(filter(var_filter, settings))
        for key in vars:
            val = settings[key]
            if not val:
                continue
            if isinstance(val, tuple):
                val = ' '.join(val)
            out.write(f'{key}="{val}"')


class arch(_base, metaclass=_register_command):
    """output the arch defined for this profile"""

    def __call__(self, namespace, out, err):
        if namespace.profile.arch is not None:
            out.write(namespace.profile.arch)


def bind_parser(parser, name):
    subparsers = parser.add_subparsers(description=f"{name} commands")
    for command in commands:
        # Split docstrings into summaries and extended docs.
        help, _, docs = command.__doc__.partition('\n')
        subparser = subparsers.add_parser(
            command.__name__.lower(),
            help=help, docs=docs)
        command().bind_to_parser(subparser)
